"""Remote project management utilities for the projectroles app"""

import json
import logging
import ssl
import urllib

from copy import deepcopy
from packaging.version import parse as parse_version
from typing import Any, Optional, Union

from django.conf import settings
from django.contrib import auth
from django.contrib.auth.models import Group
from django.core.exceptions import ObjectDoesNotExist
from django.db import transaction
from django.http import HttpRequest
from django.utils import timezone
from django.urls.base import reverse

from djangoplugins.models import Plugin

from projectroles.app_settings import AppSettingAPI, APP_SETTING_GLOBAL_DEFAULT
from projectroles.models import (
    Project,
    Role,
    RoleAssignment,
    RemoteProject,
    RemoteSite,
    SODARUser,
    SODARUserAdditionalEmail,
    SODAR_CONSTANTS,
    ROLE_RANKING,
    AppSetting,
)
from projectroles.plugins import PluginAPI


app_settings = AppSettingAPI()
logger = logging.getLogger(__name__)
plugin_api = PluginAPI()
User = auth.get_user_model()


# SODAR constants
PROJECT_TYPE_PROJECT = SODAR_CONSTANTS['PROJECT_TYPE_PROJECT']
PROJECT_TYPE_CATEGORY = SODAR_CONSTANTS['PROJECT_TYPE_CATEGORY']
PROJECT_ROLE_OWNER = SODAR_CONSTANTS['PROJECT_ROLE_OWNER']
PROJECT_ROLE_DELEGATE = SODAR_CONSTANTS['PROJECT_ROLE_DELEGATE']
PROJECT_ROLE_VIEWER = SODAR_CONSTANTS['PROJECT_ROLE_VIEWER']
SITE_MODE_TARGET = SODAR_CONSTANTS['SITE_MODE_TARGET']
SITE_MODE_SOURCE = SODAR_CONSTANTS['SITE_MODE_SOURCE']
SITE_MODE_PEER = SODAR_CONSTANTS['SITE_MODE_PEER']
REMOTE_LEVEL_NONE = SODAR_CONSTANTS['REMOTE_LEVEL_NONE']
REMOTE_LEVEL_REVOKED = SODAR_CONSTANTS['REMOTE_LEVEL_REVOKED']
REMOTE_LEVEL_VIEW_AVAIL = SODAR_CONSTANTS['REMOTE_LEVEL_VIEW_AVAIL']
REMOTE_LEVEL_READ_INFO = SODAR_CONSTANTS['REMOTE_LEVEL_READ_INFO']
REMOTE_LEVEL_READ_ROLES = SODAR_CONSTANTS['REMOTE_LEVEL_READ_ROLES']
SYSTEM_USER_GROUP = SODAR_CONSTANTS['SYSTEM_USER_GROUP']
APP_SETTING_TYPE_BOOLEAN = SODAR_CONSTANTS['APP_SETTING_TYPE_BOOLEAN']
APP_SETTING_TYPE_INTEGER = SODAR_CONSTANTS['APP_SETTING_TYPE_INTEGER']
APP_SETTING_TYPE_JSON = SODAR_CONSTANTS['APP_SETTING_TYPE_JSON']
APP_SETTING_TYPE_STRING = SODAR_CONSTANTS['APP_SETTING_TYPE_STRING']

# Local constants
APP_NAME = 'projectroles'
NO_LOCAL_USERS_MSG = 'Local users not allowed'
VERSION_2_0 = parse_version('2.0')


class RemoteProjectAPI:
    """Remote project data handling API"""

    def __init__(self):
        #: Remote data retrieved from source site
        self.remote_data = None
        #: Remote source site currently being worked with
        self.source_site = None
        #: Timeline API
        self.timeline = plugin_api.get_backend_api('timeline_backend')
        #: User for storing timeline events
        self.tl_user = None
        #: Default owner for projects
        self.default_owner = None
        #: Updated parent projects in current sync operation
        self.updated_parents = []
        #: User name/UUID lookup
        self.user_lookup = {}

    # Internal Source Site Functions -------------------------------------------

    @classmethod
    def _add_parent_categories(
        cls, sync_data: dict, category: Project, project_level: str
    ) -> dict:
        """
        Add parent categories of a project to source site sync data.

        :param sync_data: Sync data to be updated (dict)
        :param category: Project object for category
        :param project_level: Access level for the project being synced (string)
        :return: Updated sync_data (dict)
        """
        if category.parent:
            sync_data = cls._add_parent_categories(
                sync_data, category.parent, project_level
            )
        # Add if not added yet OR if a READ_ROLES project is encountered
        if str(category.sodar_uuid) not in sync_data['projects'].keys() or (
            sync_data['projects'][str(category.sodar_uuid)]['level']
            != REMOTE_LEVEL_READ_ROLES
            and project_level == REMOTE_LEVEL_READ_ROLES
        ):
            cat_data = {
                'title': category.title,
                'type': PROJECT_TYPE_CATEGORY,
                'parent_uuid': (
                    str(category.parent.sodar_uuid) if category.parent else None
                ),
                'description': category.description,
                'readme': category.readme.raw,
            }
            if project_level == REMOTE_LEVEL_READ_ROLES:
                cat_data['roles'] = {}
                cat_data['level'] = REMOTE_LEVEL_READ_ROLES
                for role_as in category.local_roles.all():
                    cat_data['roles'][str(role_as.sodar_uuid)] = {
                        'user': role_as.user.username,
                        'role': role_as.role.name,
                    }
                    sync_data = cls._add_user(sync_data, role_as.user)
            else:
                cat_data['level'] = REMOTE_LEVEL_READ_INFO
            sync_data['projects'][str(category.sodar_uuid)] = cat_data
        return sync_data

    @classmethod
    def _add_peer_site(cls, sync_data: dict, site: RemoteSite) -> dict:
        """
        Add peer site to source site sync data.

        :param sync_data: Sync data to be updated (dict)
        :param site: RemoteSite object for peer site
        :return: Updated sync_data (dict)
        """
        # Do not add sites twice
        if not sync_data['peer_sites'].get(str(site.sodar_uuid), None):
            sync_data['peer_sites'][str(site.sodar_uuid)] = {
                'name': site.name,
                'url': site.url,
                'description': site.description,
                'user_display': site.user_display,
            }
        return sync_data

    @classmethod
    def _add_user(cls, sync_data: dict, user: SODARUser) -> dict:
        """
        Add user to source site sync data.

        :param sync_data: Sync data to be updated (dict)
        :param user: SODARUser object
        :return: Updated sync_data (dict)
        """
        if user.username not in [
            u['username'] for u in sync_data['users'].values()
        ]:
            add_emails = SODARUserAdditionalEmail.objects.filter(
                user=user, verified=True
            )
            sync_data['users'][str(user.sodar_uuid)] = {
                'username': user.username,
                'name': user.name,
                'first_name': user.first_name,
                'last_name': user.last_name,
                'email': user.email,
                'additional_emails': [e.email for e in add_emails],
                'groups': [g.name for g in user.groups.all()],
                'sodar_uuid': str(user.sodar_uuid),
            }
        return sync_data

    @classmethod
    def _add_app_setting(
        cls,
        sync_data: dict,
        app_setting: AppSetting,
        all_defs: dict,
        req_version: Optional[str] = None,
    ) -> dict:
        """
        Add app setting to sync data on source site.

        :param sync_data: Sync data to be updated (dict)
        :param app_setting: AppSetting object
        :param all_defs: All settings defs (dict)
        :param req_version: Request version (string or None for default)
        :return: Updated sync_data (dict)
        """
        if app_setting.app_plugin:
            plugin_name = app_setting.app_plugin.name
        else:
            plugin_name = APP_NAME
        s_def = all_defs.get(plugin_name, {}).get(app_setting.name, {})
        if not s_def:
            # Missing def can happen if admin forgot to run cleanappsettings
            # NOTE: We don't raise this, instead log as warning
            log_msg = (
                'Definition not found for setting, skipping (run '
                'cleanappsettings)'
            )
            logger.warning(cls._get_app_setting_error(app_setting, log_msg))
            return sync_data

        # Return old-style ip_allowlist if request version <2.0
        if (
            app_setting.name == 'ip_allow_list'
            and req_version
            and parse_version(req_version) < VERSION_2_0
        ):
            s_name = 'ip_allowlist'
            s_type = APP_SETTING_TYPE_JSON
            s_value = ''
            s_value_json = (
                [v.strip() for v in app_setting.value.split(',')]
                if app_setting.value
                else []
            )
        else:
            s_name = app_setting.name
            s_type = app_setting.type
            s_value = app_setting.value
            s_value_json = app_setting.value_json

        # NOTE: Provide user_name in case of local target users
        sync_data['app_settings'][str(app_setting.sodar_uuid)] = {
            'name': s_name,
            'type': s_type,
            'value': s_value,
            'value_json': s_value_json,
            'app_plugin': (
                app_setting.app_plugin.name if app_setting.app_plugin else None
            ),
            'project_uuid': (
                str(app_setting.project.sodar_uuid)
                if app_setting.project
                else None
            ),
            'user_uuid': (
                str(app_setting.user.sodar_uuid) if app_setting.user else None
            ),
            'user_name': (
                app_setting.user.username if app_setting.user else None
            ),
            'global': s_def.global_edit,
        }
        return sync_data

    @classmethod
    def _get_app_setting_error(
        cls, app_setting: AppSetting, exception: Union[Exception, str]
    ) -> str:
        """
        Return app setting error message to be logged.

        :param app_setting: AppSetting object
        :param exception: Exception object or string
        :return: String
        """
        return (
            'Failed to add app setting "{}.settings.{}" (UUID={}): {} '
        ).format(
            app_setting.app_plugin.name if app_setting.app_plugin else APP_NAME,
            app_setting.name,
            app_setting.sodar_uuid,
            exception,
        )

    # Source Site API functions ------------------------------------------------

    def get_source_data(
        self, target_site: RemoteSite, req_version: Optional[str] = None
    ) -> dict:
        """
        Get user and project data on a source site to be synchronized into a
        target site.

        :param target_site: RemoteSite object for target site
        :param req_version: Request version (string or None for default)
        :return: Dict
        """
        sync_data = {
            'users': {},
            'projects': {},
            'peer_sites': {},
            'app_settings': {},
        }
        all_defs = app_settings.get_all_defs()

        for rp in target_site.projects.all():
            project = rp.get_project()
            # All RemoteSites which also host this project with a sufficient
            # access level
            remote_sites = [
                relation.site
                for relation in RemoteProject.objects.filter(
                    project_uuid=project.sodar_uuid
                )
                if relation.level
                in [REMOTE_LEVEL_READ_INFO, REMOTE_LEVEL_READ_ROLES]
                and relation.site != target_site  # Dont add current target site
            ]

            # Get and add app settings for project
            for a in AppSetting.objects.filter(project=project):
                try:
                    sync_data = self._add_app_setting(
                        sync_data, a, all_defs, req_version
                    )
                except Exception as ex:
                    logger.error(self._get_app_setting_error(a, ex))

            # RemoteSite data to create objects on target site
            for site in remote_sites:
                sync_data = self._add_peer_site(sync_data, site)

            project_data = {
                'level': rp.level,
                'title': project.title,
                'type': PROJECT_TYPE_PROJECT,
                'remote_sites': [str(site.sodar_uuid) for site in remote_sites],
            }
            # View available projects
            if rp.level == REMOTE_LEVEL_VIEW_AVAIL:
                project_data['available'] = True if project else False

            # Add info
            elif project and rp.level in [
                REMOTE_LEVEL_READ_INFO,
                REMOTE_LEVEL_READ_ROLES,
                REMOTE_LEVEL_REVOKED,
            ]:
                project_data['description'] = project.description
                project_data['readme'] = project.readme.raw
                # Add parent categories and inherited roles
                if project.parent:
                    sync_data = self._add_parent_categories(
                        sync_data, project.parent, rp.level
                    )
                    project_data['parent_uuid'] = str(project.parent.sodar_uuid)

            # If level is READ_ROLES or REVOKED, add roles
            if rp.level in [REMOTE_LEVEL_READ_ROLES, REMOTE_LEVEL_REVOKED]:
                project_data['roles'] = {}
                for role_as in project.local_roles.all():
                    # Omit project viewer roles if API version <2.0
                    if (
                        req_version
                        and parse_version(req_version) < VERSION_2_0
                        and role_as.role.rank
                        >= ROLE_RANKING[PROJECT_ROLE_VIEWER]
                    ):
                        logger.debug(
                            f'Skipping {role_as.role.name} role with API v1.0: '
                            f'{role_as}'
                        )
                        continue
                    # If REVOKED, only sync owner and delegate
                    if (
                        rp.level == REMOTE_LEVEL_READ_ROLES
                        or role_as.role.name
                        in ['project owner', 'project delegate']
                    ):
                        project_data['roles'][str(role_as.sodar_uuid)] = {
                            'user': role_as.user.username,
                            'role': role_as.role.name,
                        }
                        sync_data = self._add_user(sync_data, role_as.user)
            sync_data['projects'][str(rp.project_uuid)] = project_data

        # Sync USER scope settings
        for a in AppSetting.objects.filter(project=None).exclude(user=None):
            try:
                # Add user if they lack roles in remote projects (see #1593)
                self._add_user(sync_data, a.user)
                sync_data = self._add_app_setting(sync_data, a, all_defs)
            except Exception as ex:
                logger.error(self._get_app_setting_error(a, ex))
        return sync_data

    def get_remote_data(self, site: RemoteSite) -> dict:
        """
        Method for synchronisation tasks.

        :param site: Remote Site object
        :raise: Exception if some Errors occured
        :return: remote data (dict)
        """
        from projectroles.views_api import (
            SYNC_API_MEDIA_TYPE,
            SYNC_API_DEFAULT_VERSION,
        )

        api_url = site.get_url() + reverse(
            'projectroles:api_remote_get', kwargs={'secret': site.secret}
        )

        try:
            api_req = urllib.request.Request(api_url)
            api_req.add_header(
                'accept',
                f'{SYNC_API_MEDIA_TYPE}; version={SYNC_API_DEFAULT_VERSION}',
            )
            response = urllib.request.urlopen(api_req)
            remote_data = json.loads(response.read().decode('utf-8'))
        except Exception as ex:
            ex_str = str(ex)
            if (
                isinstance(ex, urllib.error.URLError)
                and isinstance(ex.reason, ssl.SSLError)
                and ex.reason.reason == 'WRONG_VERSION_NUMBER'
            ):
                ex_str = 'Most likely server cannot handle HTTPS requests.'
            if len(ex_str) >= 255:
                ex_str = ex_str[:255]
            raise Exception(ex_str)
        return remote_data

    # Internal Target Site Functions -------------------------------------------

    @staticmethod
    def _is_local_user(user_data: dict) -> bool:
        """
        Return True if user data denotes a local user.

        :param user_data: Dict
        :return: Boolean
        """
        return (
            not user_data.get('groups')
            or SYSTEM_USER_GROUP in user_data['groups']
        )

    @staticmethod
    def _update_obj(obj: Any, obj_data: dict, fields: list) -> Any:
        """
        Update object for target site sync.

        :param obj: Django database object
        :param obj_data: Object sync data (dict)
        :param fields: Fields to be updated (list)
        :return: Updated object
        """
        for f in [f for f in fields if hasattr(obj, f)]:
            setattr(obj, f, obj_data[f])
        obj.save()
        return obj

    def _check_local_categories(self, uuid: str) -> bool:
        """
        Check for local category name conflicts on a target site.

        :param uuid: Category UUID (string)
        :return: True if conflict was found (bool)
        """
        c_data = self.remote_data['projects'][uuid]
        local_cat = (
            Project.objects.filter(
                type=PROJECT_TYPE_CATEGORY, title=c_data['title']
            )
            .exclude(sodar_uuid=uuid)
            .first()
        )
        if local_cat and not local_cat.parent and not c_data['parent_uuid']:
            return True
        elif (
            local_cat
            and local_cat.parent
            and local_cat.parent.title
            == self.remote_data['projects'][c_data['parent_uuid']]['title']
        ):
            return self._check_local_categories(c_data['parent_uuid'])
        return False

    def _sync_user(self, uuid: str, user_data: dict):
        """
        Synchronize user on target site. For local users, will only update an
        existing user object. Local users must be manually created. If local
        users are not allowed, data is not synchronized.

        :param uuid: User UUID (string)
        :param user_data: User sync data (dict)
        """
        # Don't sync local users if disallowed
        allow_local = getattr(settings, 'PROJECTROLES_ALLOW_LOCAL_USERS', False)
        if self._is_local_user(user_data) and not allow_local:
            logger.info(NO_LOCAL_USERS_MSG)
            return
        # Add UUID to user_data to ensure it gets synced
        user_data['sodar_uuid'] = uuid
        # Pop additional emails
        # TODO: Simply omit this from object creation/update instead?
        add_emails = user_data.pop('additional_emails')

        # Update existing user
        try:
            user = User.objects.get(username=user_data['username'])
            updated_fields = []
            for k, v in user_data.items():
                if (
                    k != 'groups'
                    and hasattr(user, k)
                    and str(getattr(user, k)) != str(v)
                ):
                    updated_fields.append(k)

            if updated_fields:
                user = self._update_obj(user, user_data, updated_fields)
                user_data['status'] = 'updated'
                logger.info(
                    'Updated user: {} ({}): {}'.format(
                        user_data['username'], uuid, ', '.join(updated_fields)
                    )
                )

            # Check and update groups
            if sorted([g.name for g in user.groups.all()]) != sorted(
                user_data['groups']
            ):
                for g in user.groups.all():
                    if g.name not in user_data['groups']:
                        g.user_set.remove(user)
                        logger.debug(
                            f'Removed user {user.username} ({user.sodar_uuid}) '
                            f'from group "{g.name}"'
                        )
                existing_groups = [g.name for g in user.groups.all()]
                for g in user_data['groups']:
                    if g not in existing_groups:
                        group, created = Group.objects.get_or_create(name=g)
                        group.user_set.add(user)
                        logger.debug(
                            f'Added user {user.username} ({user.sodar_uuid}) '
                            f'to group "{g}"'
                        )

        # Create new user
        except User.DoesNotExist:
            if self._is_local_user(user_data):
                logger.warning(
                    'Local user "{}" not found: local users must '
                    'be manually created before they can be synced'.format(
                        user_data['username']
                    )
                )
                return
            create_values = {
                k: v for k, v in user_data.items() if k != 'groups'
            }
            user = User.objects.create(**create_values)
            user_data['status'] = 'created'
            logger.info(f'Created user: {user.username}')

            for g in user_data['groups']:
                group, created = Group.objects.get_or_create(name=g)
                group.user_set.add(user)
                logger.debug(
                    f'Added user {user.username} ({user.sodar_uuid}) to group '
                    f'"{g}"'
                )

        # Sync additional emails
        deleted_emails = SODARUserAdditionalEmail.objects.filter(
            user=user
        ).exclude(email__in=add_emails)
        email_update = False
        for e in deleted_emails:
            e.delete()
            logger.info(
                f'Deleted user {user.username} additional email "{e.email}"'
            )
            email_update = True
        for e in add_emails:
            if SODARUserAdditionalEmail.objects.filter(
                user=user, email=e
            ).exists():
                continue
            email_obj = SODARUserAdditionalEmail.objects.create(
                user=user, email=e, verified=True
            )
            logger.info(
                f'Created user {user.username} additional email '
                f'"{email_obj.email}"'
            )
            email_update = True
        # HACK: Re-add additional emails
        user_data['additional_emails'] = add_emails
        # Set user to updated if additional emails were changed
        if email_update and user_data['status'] not in ['created', 'updated']:
            user_data['status'] = 'updated'

    def _handle_user_error(
        self, error_msg: str, project: Project, role_uuid: str
    ):
        """
        Handle user sync error on target site.

        :param error_msg: Error message (string)
        :param project: Project object
        :param role_uuid: UUID of RoleAssignment object (string)
        """
        logger.error(error_msg)
        self.remote_data['projects'][str(project.sodar_uuid)]['roles'][
            role_uuid
        ]['status'] = 'error'
        self.remote_data['projects'][str(project.sodar_uuid)]['roles'][
            role_uuid
        ]['status_msg'] = error_msg

    def _handle_project_error(
        self, error_msg: str, uuid: str, project_data: dict, action: str
    ):
        """
        Handle project error on target site.

        :param error_msg: Error message (string)
        :param uuid: Project UUID (string)
        :param project_data: Project sync data (dict)
        :param action: Action resulting in error (string)
        """
        logger.error(
            '{} {} "{}" ({}): {}'.format(
                action.capitalize(),
                project_data['type'].lower(),
                project_data['title'],
                uuid,
                error_msg,
            )
        )
        self.remote_data['projects'][uuid]['status'] = 'error'
        self.remote_data['projects'][uuid]['status_msg'] = error_msg

    def _update_project(
        self, project: Project, project_data: dict, parent: Project
    ):
        """
        Update an existing project on target site.

        :param project: Project object
        :param project_data: Project sync data (dict)
        :param parent: Project object for parent category
        """
        updated_fields = []
        uuid = str(project.sodar_uuid)
        for k, v in project_data.items():
            if (
                k not in ['parent', 'sodar_uuid', 'roles', 'readme']
                and hasattr(project, k)
                and str(getattr(project, k)) != str(v)
            ):
                updated_fields.append(k)
        # README is a special case
        if project.readme.raw != project_data['readme']:
            updated_fields.append('readme')

        if updated_fields or project.parent != parent:
            project = self._update_obj(project, project_data, updated_fields)
            # Manually update parent
            if parent != project.parent:
                project.parent = parent
                project.save()
                updated_fields.append('parent')
            self.remote_data['projects'][uuid]['status'] = 'updated'
            if self.tl_user:  # Timeline
                tl_desc = (
                    'update project from remote site '
                    '"{{{}}}" ({})'.format('site', ', '.join(updated_fields))
                )
                # TODO: Add extra_data
                tl_event = self.timeline.add_event(
                    project=project,
                    app_name=APP_NAME,
                    user=self.tl_user,
                    event_name='remote_project_update',
                    description=tl_desc,
                    status_type=self.timeline.TL_STATUS_OK,
                )
                tl_event.add_object(
                    self.source_site, 'site', self.source_site.name
                )
            logger.info(
                'Updated {}: {}'.format(
                    project_data['type'].lower(),
                    ', '.join(sorted(updated_fields)),
                )
            )
        else:
            logger.debug('Nothing to update in project details')

    def _create_project(self, uuid: str, project_data: dict, parent: Project):
        """
        Create a new project on target site.

        :param uuid: Project UUID (string)
        :param project_data: Project sync data (dict)
        :param parent: Project object for parent category
        """
        # Check existing title under the same parent
        old_project = Project.objects.filter(
            parent=parent, title=project_data['title']
        ).first()
        if old_project:
            error_msg = (
                f'{old_project.type.capitalize()} with the title '
                f'"{old_project.title}" exists under the same parent, unable '
                f'to create'
            )
            self._handle_project_error(error_msg, uuid, project_data, 'create')
            return

        create_fields = ['title', 'description', 'readme']
        create_values = {
            k: v for k, v in project_data.items() if k in create_fields
        }
        create_values['type'] = project_data['type']
        create_values['parent'] = parent
        create_values['sodar_uuid'] = uuid
        project = Project.objects.create(**create_values)
        self.remote_data['projects'][uuid]['status'] = 'created'

        if self.tl_user:  # Timeline
            tl_event = self.timeline.add_event(
                project=project,
                app_name=APP_NAME,
                user=self.tl_user,
                event_name='remote_project_create',
                description='create project from remote site {site}',
                status_type=self.timeline.TL_STATUS_OK,
            )
            # TODO: Add extra_data
            tl_event.add_object(self.source_site, 'site', self.source_site.name)
        logger.info('Created {}'.format(project_data['type'].lower()))

    def _create_peer_site(self, uuid: str, site_data: dict):
        """
        Create remote peer site on target site.

        :param uuid: Remote site UUID (string)
        :param site_data: Site sync data (dict)
        """
        create_fields = ['name', 'url', 'description', 'user_display']
        create_values = {
            k: v for k, v in site_data.items() if k in create_fields
        }
        create_values['mode'] = SITE_MODE_PEER
        create_values['sodar_uuid'] = uuid
        create_values['secret'] = None  # Do not share secret of other sites
        RemoteSite.objects.create(**create_values)
        logger.info('Created Peer Site {}'.format(create_values['name']))

    def _update_peer_site(self, uuid: str, site_data: dict):
        """
        Update remote peer site on target site.

        :param uuid: Remote site UUID (string)
        :param site_data: Site sync data (dict)
        """
        site = RemoteSite.objects.filter(sodar_uuid=uuid).first()
        updated_fields = []
        for k, v in site_data.items():
            if hasattr(site, k) and str(getattr(site, k)) != str(v):
                updated_fields.append(k)
        if updated_fields:
            site = self._update_obj(site, site_data, updated_fields)
            logger.info(
                'Updated Peer Site {}: {}'.format(
                    str(site.name), ', '.join(sorted(updated_fields))
                )
            )
        else:
            logger.debug(f'Nothing to update for peer site "{uuid}"')

    def _update_roles(self, project: Project, project_data: dict):
        """
        Create or update project roles on target site.

        :param project: Project object
        :param project_data: Project sync data (dict)
        """
        # TODO: Refactor this
        uuid = str(project.sodar_uuid)
        allow_local = getattr(settings, 'PROJECTROLES_ALLOW_LOCAL_USERS', False)

        for r_uuid, r in {
            k: v for k, v in project_data['roles'].items()
        }.items():
            # Ensure the Role exists
            try:
                role = Role.objects.get(name=r['role'])
            except Role.DoesNotExist:
                error_msg = 'Role object "{}" not found (assignment {})'.format(
                    r['role'], r_uuid
                )
                self._handle_user_error(error_msg, project, r_uuid)
                continue

            # Ensure the user is valid
            local_user = self._is_local_user(
                self.remote_data['users'][self.user_lookup[r['user']]]
            )
            if (
                local_user
                and not allow_local
                and r['role'] != PROJECT_ROLE_OWNER
            ):
                error_msg = (
                    'Local user "{}" set for role "{}" but local '
                    'users are not allowed'.format(r['user'], r['role'])
                )
                self._handle_user_error(error_msg, project, r_uuid)
                continue
            # If local user, ensure they exist
            elif (
                local_user
                and allow_local
                and r['role'] != PROJECT_ROLE_OWNER
                and not User.objects.filter(username=r['user']).first()
            ):
                error_msg = (
                    'Local user "{}" not found, role of "{}" will '
                    'not be assigned'.format(r['user'], r['role'])
                )
                self._handle_user_error(error_msg, project, r_uuid)
                continue
            # Use the default owner, if owner role is for local user and local
            # users are not allowed
            if (
                local_user
                and r['role'] == PROJECT_ROLE_OWNER
                and (
                    not allow_local
                    or not User.objects.filter(username=r['user']).first()
                )
            ):
                role_user = self.default_owner
                # Notify of assigning role to default owner
                status_msg = (
                    'Local user "{}" set as owner, assigning role '
                    'to user "{}"'.format(
                        r['user'], self.default_owner.username
                    )
                )
                self.remote_data['projects'][uuid]['roles'][r_uuid][
                    'user'
                ] = self.default_owner.username
                self.remote_data['projects'][uuid]['roles'][r_uuid][
                    'status_msg'
                ] = status_msg
                logger.info(status_msg)
            else:
                role_user = User.objects.get(username=r['user'])

            # Update RoleAssignment if it exists and is changed
            old_as = RoleAssignment.objects.filter(
                project=project, user=role_user
            ).first()

            # Delete existing owner role
            if r['role'] == PROJECT_ROLE_OWNER:
                old_owner_as = project.get_owner()
                if old_owner_as and old_owner_as.user != role_user:
                    old_owner_as.delete()
                    logger.debug(
                        f'Deleted existing owner role from user '
                        f'"{old_owner_as.user.username}"'
                    )

            if old_as and old_as.role != role:
                old_as.role = role
                old_as.user = role_user
                old_as.save()
                self.remote_data['projects'][str(project.sodar_uuid)]['roles'][
                    r_uuid
                ]['status'] = 'updated'

                if self.tl_user:
                    tl_desc = (
                        'update role to "{}" for {{{}}} from site '
                        '{{{}}}'.format(role.name, 'user', 'site')
                    )
                    tl_event = self.timeline.add_event(
                        project=project,
                        app_name=APP_NAME,
                        user=self.tl_user,
                        event_name='remote_role_update',
                        description=tl_desc,
                        status_type=self.timeline.TL_STATUS_OK,
                    )
                    tl_event.add_object(role_user, 'user', role_user.username)
                    tl_event.add_object(
                        self.source_site, 'site', self.source_site.name
                    )

                logger.info(
                    f'Updated role {r_uuid}: {role_user.username} = {role.name}'
                )

            # Create a new RoleAssignment
            elif not old_as:
                role_values = {
                    'sodar_uuid': r_uuid,
                    'project': project,
                    'role': role,
                    'user': role_user,
                }
                RoleAssignment.objects.create(**role_values)
                self.remote_data['projects'][str(project.sodar_uuid)]['roles'][
                    r_uuid
                ]['status'] = 'created'

                if self.tl_user:
                    tl_desc = (
                        'add role "{}" for {{{}}} '
                        'from site {{{}}}'.format(role.name, 'user', 'site')
                    )
                    tl_event = self.timeline.add_event(
                        project=project,
                        app_name=APP_NAME,
                        user=self.tl_user,
                        event_name='remote_role_create',
                        description=tl_desc,
                        status_type=self.timeline.TL_STATUS_OK,
                    )
                    tl_event.add_object(role_user, 'user', role_user.username)
                    tl_event.add_object(
                        self.source_site, 'site', self.source_site.name
                    )
                logger.info(
                    f'Created role {r_uuid}: {role_user.username} -> '
                    f'{role.name}'
                )

    def _remove_deleted_roles(self, project: Project, project_data: dict):
        """
        Remove deleted project roles from target site.

        :param project: Project object
        :param project_data: Project sync data (dict)
        """
        timeline = plugin_api.get_backend_api('timeline_backend')
        uuid = str(project.sodar_uuid)
        current_users = [v['user'] for k, v in project_data['roles'].items()]
        deleted_roles = (
            RoleAssignment.objects.filter(project=project)
            .exclude(role__name=PROJECT_ROLE_OWNER)
            .exclude(user__username__in=current_users)
        )
        deleted_count = deleted_roles.count()

        if deleted_count > 0:
            deleted_users = sorted([r.user.username for r in deleted_roles])

            for del_as in deleted_roles:
                del_user = del_as.user
                del_role = del_as.role
                del_uuid = str(del_as.sodar_uuid)
                del_as.delete()
                self.remote_data['projects'][uuid]['roles'][del_uuid] = {
                    'user': del_user.username,
                    'role': del_role.name,
                    'status': 'deleted',
                }
                if self.tl_user:  # Timeline
                    tl_desc = (
                        'remove role "{}" from {{{}}} by site '
                        '{{{}}}'.format(del_role.name, 'user', 'site')
                    )
                    tl_event = timeline.add_event(
                        project=project,
                        app_name=APP_NAME,
                        user=self.tl_user,
                        event_name='remote_role_delete',
                        description=tl_desc,
                        status_type=timeline.TL_STATUS_OK,
                    )
                    tl_event.add_object(del_user, 'user', del_user.username)
                    tl_event.add_object(
                        self.source_site, 'site', self.source_site.name
                    )
            logger.info(
                'Deleted {} removed role{} for: {}'.format(
                    deleted_count,
                    's' if deleted_count != 1 else '',
                    ', '.join(deleted_users),
                )
            )

    def _sync_project(self, uuid: str, project_data: dict):
        """
        Synchronize a single project on target site. Create/update project, its
        parents and user roles.

        :param uuid: Project UUID (string)
        :param project_data: Project sync data (dict)
        """
        # Add/update parents if not yet handled
        if (
            project_data['parent_uuid']
            and project_data['parent_uuid'] not in self.updated_parents
        ):
            c_data = self.remote_data['projects'][project_data['parent_uuid']]
            self._sync_project(project_data['parent_uuid'], c_data)
            self.updated_parents.append(project_data['parent_uuid'])

        project = Project.objects.filter(
            type=project_data['type'], sodar_uuid=uuid
        ).first()
        parent = None
        action = 'create' if not project else 'update'
        logger.info(
            'Processing {} "{}" ({})..'.format(
                project_data['type'].lower(), project_data['title'], uuid
            )
        )

        # Get parent and ensure it exists
        if project_data['parent_uuid']:
            try:
                parent = Project.objects.get(
                    sodar_uuid=project_data['parent_uuid']
                )
            except Project.DoesNotExist:
                # Handle error
                error_msg = 'Parent {} not found'.format(
                    project_data['parent_uuid']
                )
                self._handle_project_error(
                    error_msg, uuid, project_data, action
                )
                return

        # Update/create project
        if project:
            self._update_project(project, project_data, parent)
        else:
            self._create_project(uuid, project_data, parent)
            project = Project.objects.filter(
                type=project_data['type'], sodar_uuid=uuid
            ).first()

        # Create/update a RemoteProject object
        try:
            remote_project = RemoteProject.objects.get(
                site=self.source_site, project=project
            )
            remote_project.level = project_data['level']
            remote_project.project = project
            remote_project.date_access = timezone.now()
            remote_project.save()
            remote_action = 'updated'
        except RemoteProject.DoesNotExist:
            remote_project = RemoteProject.objects.create(
                site=self.source_site,
                project_uuid=project.sodar_uuid,
                project=project,
                level=project_data['level'],
                date_access=timezone.now(),
            )
            remote_action = 'created'
        logger.debug(
            f'{remote_action.capitalize()} RemoteProject '
            f'{remote_project.sodar_uuid}'
        )

        # Skip the rest if not updating roles
        if 'level' in project_data and project_data['level'] not in [
            REMOTE_LEVEL_READ_ROLES,
            REMOTE_LEVEL_REVOKED,
        ]:
            return
        # Create/update roles
        if project_data['level'] == REMOTE_LEVEL_READ_ROLES:
            self._update_roles(project, project_data)
        # Remove deleted user roles (also for REVOKED projects)
        self._remove_deleted_roles(project, project_data)

    def _sync_peer_projects(self, uuid: str, project_data: dict):
        """
        Create RemoteProject objects on target site to represent local project
        on different peer sites.

        :param uuid: Project UUID (string)
        :param project_data: Project sync data (dict)
        """
        if project_data.get('remote_sites', None):
            for remote_site_uuid in project_data['remote_sites']:
                remote_site = RemoteSite.objects.filter(
                    sodar_uuid=remote_site_uuid
                ).first()

                try:
                    remote_project = RemoteProject.objects.get(
                        site=remote_site, project_uuid=uuid
                    )
                    remote_project.level = project_data['level']
                    remote_project.project = Project.objects.filter(
                        sodar_uuid=uuid
                    ).first()
                    remote_project.date_access = (
                        timezone.now()
                    )  # This might not be needed for Peer Projects
                    remote_action = 'updated'
                except RemoteProject.DoesNotExist:
                    remote_project = RemoteProject.objects.create(
                        site=remote_site,
                        project_uuid=uuid,
                        project=Project.objects.filter(sodar_uuid=uuid).first(),
                        level=project_data['level'],
                        date_access=timezone.now(),  # This might not be needed
                    )
                    remote_action = 'created'

            logger.debug(
                '{} Peer project {} for the following peer sites: {}'.format(
                    remote_action.capitalize(),
                    remote_project.sodar_uuid,
                    ', '.join(project_data['remote_sites']),
                )
            )
        else:
            logger.debug(f'{uuid} is not a peer project (no remote site field)')

    @classmethod
    def _remove_revoked_peers(cls, uuid: str, project_data: dict):
        """
        Remove RemoteProject objects for revoked peer projects from target site.

        :param uuid: Project UUID (string)
        :param project_data: Project sync data (dict)
        """
        removed_sites = []

        if project_data.get('remote_sites', None):
            local_peers = RemoteProject.objects.filter(
                project_uuid=uuid, site__mode=SITE_MODE_PEER
            )
            if not local_peers:
                return
            for rp in local_peers:
                if str(rp.site.sodar_uuid) not in project_data['remote_sites']:
                    removed_sites.append(rp.site.name)
                    rp.delete()

        else:  # If an empty list, remove all
            removed_sites += [
                rp.site.name
                for rp in RemoteProject.objects.filter(
                    project_uuid=uuid, site__mode=SITE_MODE_PEER
                )
            ]
            RemoteProject.objects.filter(
                project_uuid=uuid, site__mode=SITE_MODE_PEER
            ).delete()

        if len(removed_sites) > 0:
            logger.debug(
                'Removed peer project(s) for the following sites: {}'.format(
                    ', '.join(removed_sites)
                )
            )

    def _sync_app_setting(self, uuid: str, set_data: dict):
        """
        Create or update an AppSetting on a target site.

        :param uuid: App setting UUID (string)
        :param set_data: App setting data (dict)
        """
        ad = deepcopy(set_data)
        app_plugin = None
        user = None
        user_name = ad.get('user_name')
        user_uuid = ad.get('user_uuid')
        project = None
        obj = None
        skip_msg = 'Skipping setting "{}": '.format(ad['name'])

        # Get app plugin (skip the rest if not found on target server)
        if ad['app_plugin']:
            app_plugin = Plugin.objects.filter(name=ad['app_plugin']).first()
            if not app_plugin:
                logger.debug(
                    skip_msg + 'App plugin not found with name '
                    '"{}"'.format(ad['app_plugin'])
                )
                return
        if user_name:
            local_user = self._is_local_user(
                self.remote_data['users'][user_uuid]
            )
            allow_local = getattr(
                settings, 'PROJECTROLES_ALLOW_LOCAL_USERS', False
            )
            if local_user and not allow_local:
                logger.info(skip_msg + NO_LOCAL_USERS_MSG)
                return
            user = User.objects.filter(sodar_uuid=user_uuid).first()
            # User may not be found if e.g. local users allowed but not created
            if not user:
                logger.info(
                    'Skipping setting "{}": User not found (user_name={}; '
                    'user_uuid={})'.format(ad['name'], user_name, user_uuid)
                )
                return
        if ad['project_uuid']:
            project = Project.objects.get(sodar_uuid=ad['project_uuid'])

        try:
            obj = AppSetting.objects.get(
                app_plugin=app_plugin,
                name=ad['name'],
                project=project,
                user=user,
            )
            # Keep target instance of local app setting if available
            if not ad.get('global', APP_SETTING_GLOBAL_DEFAULT):
                logger.info('Keeping local setting {}'.format(ad['name']))
                set_data['status'] = 'skipped'
                return
            # Skip if value is identical
            if app_settings.compare_value(
                obj,
                (
                    ad['value_json']
                    if obj.type == APP_SETTING_TYPE_JSON
                    else ad['value']
                ),
            ):
                logger.info(f'Skipping setting {obj}: value unchanged')
                set_data['status'] = 'skipped'
                return
            # Else update existing value
            action_str = 'updating'
        except ObjectDoesNotExist:
            action_str = 'creating'

        # Remove keys that are not available in the model
        ad.pop('global', None)
        ad.pop('local', None)  # Deprecated in v1.0, remove just in case
        ad.pop('project_uuid', None)
        ad.pop('user_name', None)
        ad.pop('user_uuid', None)
        # Add keys required for the model
        ad['project'] = project
        ad['user'] = user
        ad['sodar_uuid'] = uuid
        if app_plugin:
            ad['app_plugin'] = app_plugin

        logger.info(f'{action_str.capitalize()} setting {obj}')
        logger.debug(f'Setting data: {ad}')
        if action_str == 'updating' and obj:  # Update existing setting object
            for k, v in ad.items():
                setattr(obj, k, v)
        else:  # Create new setting object
            obj = AppSetting(**ad)
        obj.save()
        set_data['status'] = action_str.replace('ing', 'ed')

    # Target Site API functions ------------------------------------------------

    @transaction.atomic
    def sync_remote_data(
        self,
        site: RemoteSite,
        remote_data: dict,
        request: Optional[HttpRequest] = None,
    ) -> dict:
        """
        Synchronize remote user and project data into the local Django database
        on a target site and return information of additions.

        :param site: RemoteSite object for the source site
        :param remote_data: Data returned by get_source_data() on the source
                            site (dict)
        :param request: Request object (optional)
        :return: Dict with updated remote_data
        :raise: ValueError if user from PROJECTROLES_DEFAULT_ADMIN is not found
        """
        self.source_site = site
        self.remote_data = remote_data
        self.updated_parents = []

        # Get default owner if remote projects have a local owner
        try:
            self.default_owner = User.objects.get(
                username=settings.PROJECTROLES_DEFAULT_ADMIN
            )
        except User.DoesNotExist:
            error_msg = (
                f'Local user "{settings.PROJECTROLES_DEFAULT_ADMIN}" defined '
                f'in PROJECTROLES_DEFAULT_ADMIN not found'
            )
            logger.error(error_msg)
            raise ValueError(error_msg)

        # Check for name conflicts in local categories
        for k, v in self.remote_data['projects'].items():
            if (
                v['type'] == PROJECT_TYPE_PROJECT
                and v['parent_uuid']
                and self._check_local_categories(v['parent_uuid'])
            ):
                error_msg = (
                    'Remote sync cancelled: Existing local category '
                    'structure on target site'
                )
                logger.error(error_msg)
                raise ValueError(error_msg)

        # Set up timeline user
        if self.timeline:
            self.tl_user = request.user if request else self.default_owner

        logger.info(f'Synchronizing data from "{site.name}"..')
        # Return unchanged data if no projects with READ_ROLES are included
        if not {
            k: v
            for k, v in self.remote_data['projects'].items()
            if v['type'] == PROJECT_TYPE_PROJECT
            and v['level'] in [REMOTE_LEVEL_READ_ROLES, REMOTE_LEVEL_REVOKED]
        }.values():
            logger.info(
                'No READ_ROLES or REVOKED access set, nothing to synchronize'
            )
            return self.remote_data

        # Peer Sites
        logger.info('Synchronizing peer sites..')
        if self.remote_data.get('peer_sites', None):
            for remote_site_uuid, site_data in self.remote_data[
                'peer_sites'
            ].items():
                # Create RemoteSite Objects if not yet there
                remote_site = RemoteSite.objects.filter(
                    sodar_uuid=remote_site_uuid
                ).first()
                if remote_site:
                    self._update_peer_site(remote_site_uuid, site_data)
                else:
                    self._create_peer_site(remote_site_uuid, site_data)
            logger.info('Peer site sync OK')
        else:
            logger.info('No peer sites to sync')

        # Users
        logger.info('Synchronizing users..')
        # NOTE: Add all users here, only update local users within _sync_user()
        for u_uuid, u_data in self.remote_data['users'].items():
            self._sync_user(u_uuid, u_data)
            # HACK: Populate user lookup
            self.user_lookup[u_data['username']] = u_uuid
        logger.info('User sync OK')

        # Categories and Projects
        logger.info('Synchronizing projects..')
        for p_uuid, p_data in {
            k: v
            for k, v in self.remote_data['projects'].items()
            if v['type'] == PROJECT_TYPE_PROJECT
            and v['level'] in [REMOTE_LEVEL_READ_ROLES, REMOTE_LEVEL_REVOKED]
        }.items():
            self._sync_project(p_uuid, p_data)
            self._sync_peer_projects(p_uuid, p_data)
            self._remove_revoked_peers(p_uuid, p_data)

        # App Settings
        logger.info('Synchronizing app settings..')
        for a_uuid, a_data in self.remote_data['app_settings'].items():
            try:
                self._sync_app_setting(a_uuid, a_data)
            except Exception as ex:
                logger.error(
                    'Failed to set app setting "{}.setting.{}" ({}): {}'.format(
                        (
                            a_data['app_plugin']
                            if a_data['app_plugin']
                            else APP_NAME
                        ),
                        a_data['name'],
                        a_uuid,
                        ex,
                    )
                )
                if settings.DEBUG:
                    raise ex
        logger.info('Synchronization OK')
        return self.remote_data



"""UI views for the adminalerts app"""

import logging

from django.conf import settings
from django.contrib import messages
from django.contrib.auth import get_user_model
from django.shortcuts import redirect
from django.urls import reverse
from django.views.generic import (
    DetailView,
    UpdateView,
    CreateView,
    DeleteView,
    ListView,
)
from django.views.generic.edit import ModelFormMixin

# Projectroles dependency
from projectroles.app_settings import AppSettingAPI
from projectroles.email import get_user_addr, send_generic_mail
from projectroles.views import (
    LoggedInPermissionMixin,
    HTTPRefererMixin,
    CurrentUserFormMixin,
    InvalidFormMixin,
)

from adminalerts.forms import AdminAlertForm
from adminalerts.models import AdminAlert


app_settings = AppSettingAPI()
logger = logging.getLogger(__name__)
User = get_user_model()


# Local constants
APP_NAME = 'adminalerts'
DEFAULT_PAGINATION = 15
EMAIL_SUBJECT = '{state} admin alert: {message}'
EMAIL_BODY = r'''
An admin alert has been {action}d by {issuer}:

{message}
'''.lstrip()
EMAIL_BODY_DESCRIPTION = r'''
Additional details:
----------------------------------------
{description}
----------------------------------------
'''


# Listing/details views --------------------------------------------------------


class AdminAlertListView(LoggedInPermissionMixin, ListView):
    """Alert list view"""

    permission_required = 'adminalerts.create_alert'
    template_name = 'adminalerts/alert_list.html'
    model = AdminAlert
    paginate_by = getattr(
        settings, 'ADMINALERTS_PAGINATION', DEFAULT_PAGINATION
    )
    slug_url_kwarg = 'adminalert'
    slug_field = 'sodar_uuid'

    def get_queryset(self):
        return AdminAlert.objects.all().order_by('-pk')


class AdminAlertDetailView(
    LoggedInPermissionMixin, HTTPRefererMixin, DetailView
):
    """Alert detail view"""

    permission_required = 'adminalerts.view_alert'
    template_name = 'adminalerts/alert_detail.html'
    model = AdminAlert
    slug_url_kwarg = 'adminalert'
    slug_field = 'sodar_uuid'


# Modification views -----------------------------------------------------------


class AdminAlertModifyMixin(ModelFormMixin):
    """Common modification methods for AdminAlert create/update views"""

    @classmethod
    def _get_email_recipients(cls, alert: AdminAlert) -> list[str]:
        """Return list of email addresses for alert email recipients"""
        ret = []
        for u in User.objects.exclude(is_active=False):
            if not app_settings.get(APP_NAME, 'notify_email_alert', user=u):
                continue
            user_emails = get_user_addr(u)
            if not user_emails:
                logger.warning(f'No emails set for user: {u.username}')
                continue
            ret += [e for e in user_emails if e not in ret]
        return sorted(ret)

    def _send_email(self, alert: AdminAlert, action: str) -> int:
        """
        Send email alerts to all users except for the alert issuer.

        :param alert: AdminAlert object
        :param action: "create" or "update" (string)
        :return: Amount of sent mail (int)
        """
        subject = EMAIL_SUBJECT.format(
            state='New' if action == 'create' else 'Updated',
            message=alert.message,
        )
        body = EMAIL_BODY.format(
            action=action,
            issuer=alert.user.get_display_name(inc_user=True),
            message=alert.message,
        )
        if alert.description:
            body += EMAIL_BODY_DESCRIPTION.format(
                description=alert.description.raw
            )
        recipients = self._get_email_recipients(alert)
        # NOTE: Recipients go under bcc
        # NOTE: If we have no recipients in bcc we cancel sending
        if len(recipients) == 0:
            return 0
        return send_generic_mail(
            subject,
            body,
            [settings.EMAIL_SENDER],
            self.request,
            reply_to=None,
            bcc=recipients,
        )

    def form_valid(self, form):
        form_action = 'update' if self.object else 'create'
        self.object = form.save()
        email_count = 0
        email_msg_suffix = ''
        if (
            form.cleaned_data['send_email']
            and self.object.active
            and settings.PROJECTROLES_SEND_EMAIL
        ):
            email_count = self._send_email(form.instance, form_action)
        if email_count > 0:
            email_msg_suffix = ', email sent to site users'
        messages.success(
            self.request, f'Alert {form_action}d{email_msg_suffix}.'
        )
        return redirect(reverse('adminalerts:list'))


class AdminAlertCreateView(
    LoggedInPermissionMixin,
    AdminAlertModifyMixin,
    HTTPRefererMixin,
    CurrentUserFormMixin,
    InvalidFormMixin,
    CreateView,
):
    """AdminAlert creation view"""

    model = AdminAlert
    form_class = AdminAlertForm
    permission_required = 'adminalerts.create_alert'


class AdminAlertUpdateView(
    LoggedInPermissionMixin,
    AdminAlertModifyMixin,
    HTTPRefererMixin,
    CurrentUserFormMixin,
    InvalidFormMixin,
    UpdateView,
):
    """AdminAlert updating view"""

    model = AdminAlert
    form_class = AdminAlertForm
    permission_required = 'adminalerts.update_alert'
    slug_url_kwarg = 'adminalert'
    slug_field = 'sodar_uuid'


class AdminAlertDeleteView(
    LoggedInPermissionMixin, HTTPRefererMixin, DeleteView
):
    """AdminAlert deletion view"""

    model = AdminAlert
    permission_required = 'adminalerts.update_alert'
    slug_url_kwarg = 'adminalert'
    slug_field = 'sodar_uuid'

    def get_success_url(self):
        """Override for redirecting alert list view with message"""
        messages.success(self.request, 'Alert deleted.')
        return reverse('adminalerts:list')



"""Backend API tests for the appalerts app"""

from django.forms.models import model_to_dict
from django.urls import reverse

from djangoplugins.models import Plugin
from test_plus.test import TestCase

# Projectroles dependency
from projectroles.models import SODAR_CONSTANTS
from projectroles.plugins import PluginAPI
from projectroles.tests.test_models import ProjectMixin

from appalerts.models import AppAlert
from appalerts.tests.test_models import (
    AppAlertMixin,
    ALERT_NAME,
    ALERT_MSG,
    ALERT_LEVEL,
)


plugin_api = PluginAPI()


# SODAR constants
PROJECT_TYPE_PROJECT = SODAR_CONSTANTS['PROJECT_TYPE_PROJECT']

# Local constants
APP_NAME_FF = 'filesfolders'


class TestAppAlertAPI(AppAlertMixin, ProjectMixin, TestCase):
    """Base class for appalerts backend API testing"""

    def setUp(self):
        # Create user
        self.user = self.make_user('user')
        # Create superuser
        self.superuser = self.make_user('superuser')
        self.superuser.is_superuser = True
        self.project = self.make_project(
            title='TestProject', type=PROJECT_TYPE_PROJECT, parent=None
        )
        self.project_url = reverse(
            'projectroles:detail',
            kwargs={'project': str(self.project.sodar_uuid)},
        )
        # Get backend
        self.app_alerts = plugin_api.get_backend_api('appalerts_backend')

    def test_get_model(self):
        """Test get_model()"""
        self.assertEqual(self.app_alerts.get_model(), AppAlert)

    def test_add_alert(self):
        """Test alert addition with a plugin"""
        self.assertEqual(AppAlert.objects.count(), 0)
        alert = self.app_alerts.add_alert(
            app_name=APP_NAME_FF,
            alert_name=ALERT_NAME,
            user=self.user,
            message=ALERT_MSG,
            level=ALERT_LEVEL,
            url=self.project_url,
            project=self.project,
        )
        self.assertEqual(AppAlert.objects.count(), 1)
        expected = {
            'id': alert.pk,
            'app_plugin': Plugin.objects.get(name=APP_NAME_FF).pk,
            'alert_name': ALERT_NAME,
            'user': self.user.pk,
            'message': ALERT_MSG,
            'level': ALERT_LEVEL,
            'active': True,
            'url': '/project/' + str(self.project.sodar_uuid),
            'project': self.project.pk,
            'sodar_uuid': alert.sodar_uuid,
        }
        model_dict = model_to_dict(alert)
        self.assertEqual(model_dict, expected)

    def test_add_alerts(self):
        """Test add_alerts()"""
        self.assertEqual(AppAlert.objects.count(), 0)
        self.app_alerts.add_alerts(
            app_name='timeline',
            alert_name=ALERT_NAME,
            users=[self.user, self.superuser],
            message=ALERT_MSG,
            level=ALERT_LEVEL,
            url=self.project_url,
            project=self.project,
        )
        self.assertEqual(AppAlert.objects.count(), 2)

    def test_add_alert_projectroles(self):
        """Test alert addition for projectroles"""
        self.assertEqual(AppAlert.objects.count(), 0)
        alert = self.app_alerts.add_alert(
            app_name='projectroles',
            alert_name=ALERT_NAME,
            user=self.user,
            message=ALERT_MSG,
            level=ALERT_LEVEL,
            url=self.project_url,
            project=self.project,
        )
        self.assertEqual(AppAlert.objects.count(), 1)
        self.assertIsNone(alert.app_plugin)

    def test_add_alert_invalid_plugin(self):
        """Test alert addition with an invalid plugin name"""
        with self.assertRaises(ValueError):
            self.app_alerts.add_alert(
                app_name='Not a valid plugin name',
                alert_name=ALERT_NAME,
                user=self.user,
                message=ALERT_MSG,
                level=ALERT_LEVEL,
                url=self.project_url,
                project=self.project,
            )

    def test_add_alert_invalid_level(self):
        """Test alert addition with an invalid level"""
        with self.assertRaises(ValueError):
            self.app_alerts.add_alert(
                app_name=APP_NAME_FF,
                alert_name=ALERT_NAME,
                user=self.user,
                message=ALERT_MSG,
                level='Not a valid level',
                url=self.project_url,
                project=self.project,
            )



"""Models for the projectroles app"""

import logging
import uuid

from datetime import datetime
from typing import Any, Optional, Union

from django.apps import apps
from django.conf import settings
from django.contrib.auth.models import AbstractUser, Group
from django.contrib.postgres.fields import ArrayField
from django.core.exceptions import ValidationError
from django.db import models
from django.db.models import Q, QuerySet
from django.http import HttpRequest
from django.urls import reverse
from django.utils import timezone
from django.utils.translation import gettext_lazy as _

from djangoplugins.models import Plugin
from markupfield.fields import MarkupField

from projectroles.constants import get_sodar_constants


AUTH_USER_MODEL = getattr(settings, 'AUTH_USER_MODEL', 'auth.User')
logger = logging.getLogger(__name__)


# SODAR constants
SODAR_CONSTANTS = get_sodar_constants()
PROJECT_TYPE_PROJECT = SODAR_CONSTANTS['PROJECT_TYPE_PROJECT']
PROJECT_TYPE_CATEGORY = SODAR_CONSTANTS['PROJECT_TYPE_CATEGORY']
PROJECT_ROLE_OWNER = SODAR_CONSTANTS['PROJECT_ROLE_OWNER']
PROJECT_ROLE_DELEGATE = SODAR_CONSTANTS['PROJECT_ROLE_DELEGATE']
PROJECT_ROLE_CONTRIBUTOR = SODAR_CONSTANTS['PROJECT_ROLE_CONTRIBUTOR']
PROJECT_ROLE_GUEST = SODAR_CONSTANTS['PROJECT_ROLE_GUEST']
PROJECT_ROLE_VIEWER = SODAR_CONSTANTS['PROJECT_ROLE_VIEWER']
PROJECT_ROLE_FINDER = SODAR_CONSTANTS['PROJECT_ROLE_FINDER']
APP_SETTING_SCOPE_SITE = SODAR_CONSTANTS['APP_SETTING_SCOPE_SITE']
APP_SETTING_TYPE_BOOLEAN = SODAR_CONSTANTS['APP_SETTING_TYPE_BOOLEAN']
APP_SETTING_TYPE_INTEGER = SODAR_CONSTANTS['APP_SETTING_TYPE_INTEGER']
APP_SETTING_TYPE_JSON = SODAR_CONSTANTS['APP_SETTING_TYPE_JSON']
APP_SETTING_TYPE_STRING = SODAR_CONSTANTS['APP_SETTING_TYPE_STRING']
SYSTEM_USER_GROUP = SODAR_CONSTANTS['SYSTEM_USER_GROUP']
AUTH_TYPE_LOCAL = SODAR_CONSTANTS['AUTH_TYPE_LOCAL']
AUTH_TYPE_LDAP = SODAR_CONSTANTS['AUTH_TYPE_LDAP']
AUTH_TYPE_OIDC = SODAR_CONSTANTS['AUTH_TYPE_OIDC']

# Local constants
APP_NAME = 'projectroles'
ROLE_RANKING = {
    PROJECT_ROLE_OWNER: 10,
    PROJECT_ROLE_DELEGATE: 20,
    PROJECT_ROLE_CONTRIBUTOR: 30,
    PROJECT_ROLE_GUEST: 40,
    PROJECT_ROLE_VIEWER: 45,
    PROJECT_ROLE_FINDER: 50,
}
PROJECT_TYPE_CHOICES = [('CATEGORY', 'Category'), ('PROJECT', 'Project')]
APP_SETTING_TYPES = [
    APP_SETTING_TYPE_BOOLEAN,
    APP_SETTING_TYPE_INTEGER,
    APP_SETTING_TYPE_STRING,
    APP_SETTING_TYPE_JSON,
]
PROJECT_SEARCH_TYPES = ['project']
PROJECT_TAG_STARRED = 'STARRED'
CAT_DELIMITER = ' / '
CAT_DELIMITER_ERROR_MSG = f'String "{CAT_DELIMITER}" is not allowed in title'
ROLE_PROJECT_TYPE_ERROR_MSG = (
    'Invalid project type "{project_type}" for ' 'role "{role_name}"'
)
CAT_PUBLIC_ACCESS_MSG = 'Public access is not allowed for categories'
ADD_EMAIL_ALREADY_SET_MSG = 'Email already set as {email_type} email for user'
REMOTE_PROJECT_UNIQUE_MSG = (
    'RemoteProject with the same project UUID and site anready exists'
)
AUTH_PROVIDER_OIDC = 'oidc'
SET_PUBLIC_DEPRECATE_MSG = (
    'Project.set_public() has been deprecated and will be removed in SODAR '
    'Core v1.3. Use set_public_access() instead'
)


# Project ----------------------------------------------------------------------


class ProjectManager(models.Manager):
    """Manager for custom table-level Project queries"""

    def find(
        self,
        search_terms: list[str],
        keywords: Optional[dict] = None,
        project_type: Optional[str] = None,
    ) -> QuerySet:
        """
        Return projects with a partial match in full title or, including titles
        of parent Project objects, or the description of the current object.
        Restrict to project type if project_type is set.

        :param search_terms: Search terms (list)
        :param keywords: Optional search keywords as key/value pairs (dict)
        :param project_type: Project type or None
        :return: QuerySet of Project objects
        """
        projects = super().get_queryset().order_by('title')
        if project_type:
            projects = projects.filter(type=project_type)
        term_query = Q()
        for t in search_terms:
            term_query.add(Q(full_title__icontains=t), Q.OR)
            term_query.add(Q(description__icontains=t), Q.OR)
            # Only add UUID if term is valid UUID
            try:
                uuid.UUID(t)
                term_query.add(Q(sodar_uuid=t), Q.OR)
            except ValueError:
                pass
        return projects.filter(term_query).order_by('full_title')


class Project(models.Model):
    """
    A SODAR project. Can have one parent category in case of nested
    projects. The project must be of a specific type, of which "CATEGORY" and
    "PROJECT" are currently implemented. "CATEGORY" projects are used as
    containers for other projects.
    """

    #: Project title
    title = models.CharField(
        max_length=255, unique=False, help_text='Project title'
    )

    #: Type of project ("CATEGORY", "PROJECT")
    type = models.CharField(
        max_length=64,
        choices=PROJECT_TYPE_CHOICES,
        default=PROJECT_TYPE_PROJECT,
        help_text='Type of project ("CATEGORY", "PROJECT")',
    )

    #: Parent category if nested, otherwise null
    parent = models.ForeignKey(
        'self',
        blank=True,
        null=True,
        related_name='children',
        help_text='Parent category if nested',
        on_delete=models.CASCADE,
    )

    #: Short project description
    description = models.CharField(
        max_length=512,
        unique=False,
        blank=True,
        null=True,
        help_text='Short project description',
    )

    #: Project README (optional, supports markdown)
    readme = MarkupField(
        null=True,
        blank=True,
        markup_type='markdown',
        help_text='Project README (optional, supports markdown)',
    )

    #: Public guest access (DEPRECATED, to be removed in v1.3. See #1703)
    public_guest_access = models.BooleanField(
        default=False,
        help_text='Allow public guest access for the project, also including '
        'unauthenticated users if allowed on the site',
    )

    #: Public read-only access
    public_access = models.ForeignKey(
        'Role',
        blank=True,
        null=True,
        related_name='public_access',
        help_text='Allow public access of specific read-only access level for '
        'project. Includes access for unauthenticated users if allowed on '
        'site.',
        on_delete=models.CASCADE,
    )

    #: Project is archived (read-only)
    archive = models.BooleanField(
        default=False,
        help_text='Project is archived (read-only)',
    )

    #: Full project title with parent path (auto-generated)
    full_title = models.CharField(
        max_length=4096,
        null=True,
        help_text='Full project title with parent path (auto-generated)',
    )

    #: Whether project has children with public access (auto-generated)
    has_public_children = models.BooleanField(
        default=False,
        help_text='Whether project has children with public access '
        '(auto-generated)',
    )

    #: Project SODAR UUID
    sodar_uuid = models.UUIDField(
        default=uuid.uuid4, unique=True, help_text='Project SODAR UUID'
    )

    # Set manager for custom queries
    objects = ProjectManager()

    class Meta:
        unique_together = ('title', 'parent')
        ordering = ['parent__title', 'title']

    def __str__(self):
        return self.full_title

    def __repr__(self):
        values = (
            self.title,
            self.type,
            self.parent.title if self.parent else None,
        )
        return 'Project({})'.format(', '.join(repr(v) for v in values))

    def _validate_parent(self):
        """
        Validate parent value to ensure project can't be set as its own parent.
        """
        if self.parent == self:
            raise ValidationError('Project can not be set as its own parent')

    def _validate_parent_type(self):
        """Validate parent value to ensure parent can not be a project"""
        if self.parent and self.parent.is_project():
            raise ValidationError(
                'Subprojects are only allowed within categories'
            )

    def _validate_public_access(self):
        """
        Validate public access to ensure it is not set on categories and
        contains one of accepted values.

        NOTE: Does not prevent saving but forces the value to be None if in
              category, see issue #1404.
        """
        if self.is_category() and (
            self.public_access or self.public_guest_access
        ):
            logger.warning(CAT_PUBLIC_ACCESS_MSG + ', setting to None')
            self.public_access = None
            self.public_guest_access = False  # DEPRECATED
        if self.public_access and self.public_access.name not in [
            PROJECT_ROLE_GUEST,
            PROJECT_ROLE_VIEWER,
        ]:
            raise ValidationError(
                f'Invalid role for public_access: {self.public_access.name}'
            )

    def _validate_title(self):
        """
        Validate title against parent title to ensure they don't equal parent.
        """
        if self.parent and self.title == self.parent.title:
            raise ValidationError('Project and parent titles can not be equal')
        if (
            CAT_DELIMITER in self.title
            or self.title.startswith(CAT_DELIMITER.strip())
            or self.title.endswith(CAT_DELIMITER.strip())
        ):
            raise ValidationError(CAT_DELIMITER_ERROR_MSG)

    def _validate_archive(self):
        """
        Validate archive status against project type to ensure archiving is only
        applied to projects.
        """
        if self.archive and self.is_category():
            raise ValidationError(
                'Archiving a category is not currently supported'
            )

    def save(self, *args, **kwargs):
        """Custom validation and field populating for Project"""
        self._validate_parent()
        self._validate_title()
        self._validate_parent_type()
        self._validate_public_access()
        self._validate_archive()
        # Update full title of self and children
        self.full_title = self._get_full_title()
        # TODO: Save with commit=False with other args to avoid double save()?
        super().save(*args, **kwargs)
        if self.is_category():
            for child in self.children.all():
                child.save()
        # Update public children
        # NOTE: Parents will be updated in ProjectModifyMixin.modify_project()
        if self._has_public_children():
            self.has_public_children = True
            super().save(*args, **kwargs)

    def get_absolute_url(self) -> str:
        return reverse(
            'projectroles:detail', kwargs={'project': self.sodar_uuid}
        )

    # Internal helpers

    def _get_full_title(self) -> str:
        """Return full title of project with path"""
        parents = self.get_parents()
        ret = (
            CAT_DELIMITER.join([p.title for p in parents]) + CAT_DELIMITER
            if parents
            else ''
        )
        return ret + self.title

    def _has_public_children(self) -> bool:
        """
        Return True if the project has any children with public read-only
        access.
        """
        if self.is_project():
            return False
        for child in self.get_children():
            if child.public_access:
                return True
            ret = child._has_public_children()
            if ret:
                return True
        return False

    def _update_public_children(self):
        """Update has_public_children for this project's parents"""
        if self.parent:
            parent = self.parent
            public_found = False
            while parent:
                if public_found:
                    parent.has_public_children = True
                else:
                    parent.has_public_children = parent._has_public_children()
                parent.save()
                if not public_found and parent.has_public_children:
                    public_found = True
                parent = parent.parent

    # Custom row-level functions

    def is_project(self) -> bool:
        """Return True if project is of type PROJECT_TYPE_PROJECT"""
        return self.type == PROJECT_TYPE_PROJECT

    def is_category(self) -> bool:
        """Return True if project is of type PROJECT_TYPE_CATEGORY"""
        return self.type == PROJECT_TYPE_CATEGORY

    def get_parents(self) -> list:
        """
        Return a list of parent projects in inheritance order.

        :return: List of Project objects
        """
        if not self.parent:
            return []
        ret = []
        parent = self.parent
        while parent:
            ret.append(parent)
            parent = parent.parent
        ret.reverse()
        return ret

    def get_children(self, flat: bool = False) -> QuerySet:
        """
        Return child objects for a Category, sorted by full title.

        :param flat: Return all children recursively as a flat list (bool)
        :return: QuerySet
        """
        if self.is_project():
            return Project.objects.none()
        if flat:
            return Project.objects.filter(
                full_title__startswith=self.full_title + CAT_DELIMITER
            ).order_by('full_title')
        return self.children.all().order_by('title')

    def get_depth(self) -> int:
        """Return depth of project in the project tree structure (root=0)"""
        return len(self.full_title.split(CAT_DELIMITER)) - 1

    def get_public_access_name(self) -> Optional[str]:
        """
        Return public_access role name or None if public_access is not set.

        :return: String or None
        """
        return self.public_access.name if self.public_access else None

    def get_role(
        self, user: 'SODARUser', inherited_only: bool = False
    ) -> Optional['RoleAssignment']:
        """
        Return the currently active role for user, or None if not available.
        Returns the highest ranked role including inherited roles. In case of
        multiple roles of the same level in the hierarchy, the lowest one is
        returned.

        :param user: User object
        :param inherited_only: Only return an inherited role if True
                               (boolean, default=False)
        :return: RoleAssignment object or None
        """
        if not user or user.is_anonymous:
            return None
        projects = [self] if not inherited_only else []
        projects += list(self.get_parents())
        return (
            RoleAssignment.objects.filter(
                project__in=projects,
                user=user,
                role__project_types__contains=[self.type],
            )
            .order_by('role__rank', '-project__full_title')
            .first()
        )

    def get_roles(
        self,
        user: Optional['SODARUser'] = None,
        inherited: bool = True,
        inherited_only: bool = False,
        min_rank: Optional[int] = None,
        max_rank: Optional[int] = None,
    ) -> list['RoleAssignment']:
        """
        Return project role assignments.

        :param user: Limit to user (User object or None)
        :param inherited: Include inherited roles (bool)
        :param inherited_only: Return only inherited roles (bool)
        :param min_rank: Limit roles to minimum rank (integer or None)
        :param max_rank: Limit roles to maximum rank (integer or None)
        :return: List of RoleAssignment objects
        :raise: ValueError If inheritance arguments conflict
        """
        if not inherited and inherited_only:
            raise ValueError(
                'Inherited set False and inherited_only set True, No results '
                'can be returned'
            )
        projects = [] if inherited_only else [self]
        # NOTE: We have to get inherited roles to exclude overridden ones
        parents = list(self.get_parents())
        projects += parents
        q_kwargs = {
            'project__in': projects,
            'role__project_types__contains': [self.type],
        }
        if user and user.is_authenticated:
            q_kwargs['user'] = user
        roles = RoleAssignment.objects.filter(**q_kwargs).order_by(
            '-project__full_title', 'role__name', 'user'
        )
        user_roles = {}
        for a in roles:
            u = a.user
            rank_ok = (not min_rank or a.role.rank >= min_rank) and (
                not max_rank or a.role.rank <= max_rank
            )
            # Local role (always returned first if it exists)
            if a.project == self and not inherited_only and rank_ok:
                user_roles[u] = a
            # Inherited role of higher rank
            elif (
                inherited
                and a.project in parents
                and (
                    u not in user_roles or a.role.rank < user_roles[u].role.rank
                )
                and rank_ok
            ):
                user_roles[u] = a
            # Pop overridden role if in list
            elif (
                a.project in parents
                and u in user_roles
                and a.role.rank < user_roles[u].role.rank
            ):
                user_roles.pop(u, None)
        return list(user_roles.values())

    def get_roles_by_rank(
        self,
        role_name: str,
        inherited: bool = True,
        inherited_only: bool = False,
    ) -> list['RoleAssignment']:
        """
        Return RoleAssignments for specific role name. Will also include custom
        roles of identical rank once role customization is implemented (see
        issue #288).

        :param role_name: Name of role (string)
        :param inherited: Include inherited roles (bool)
        :param inherited_only: Return only inherited roles (bool)
        :return: List
        """
        if role_name not in ROLE_RANKING:
            role = Role.objects.filter(name=role_name).first()
            if not role:
                raise ValueError(f'Unknown role "{role_name}"')
            rank = role.rank
        else:
            rank = ROLE_RANKING[role_name]
        return self.get_roles(
            inherited=inherited,
            inherited_only=inherited_only,
            min_rank=rank,
            max_rank=rank,
        )

    def get_owner(self) -> Optional['RoleAssignment']:
        """
        Return RoleAssignment for local (non-inherited) owner or None if not
        set.

        :return: RoleAssignment or None
        """
        return self.local_roles.filter(role__name=PROJECT_ROLE_OWNER).first()

    def get_owners(
        self, inherited: bool = True, inherited_only: bool = False
    ) -> list['RoleAssignment']:
        """
        Return RoleAssignments for project owner as well as possible inherited
        owners from parent categories.

        :param inherited: Include inherited roles (bool)
        :param inherited_only: Return only inherited roles (bool)
        :return: List
        """
        rank = ROLE_RANKING[PROJECT_ROLE_OWNER]
        return self.get_roles(
            inherited=inherited,
            inherited_only=inherited_only,
            min_rank=rank,
            max_rank=rank,
        )

    def get_delegates(
        self, inherited: bool = True, inherited_only: bool = False
    ) -> list['RoleAssignment']:
        """
        Return RoleAssignments for delegates including inherited delegates from
        parent categories.

        :param inherited: Include inherited roles (bool)
        :param inherited_only: Return only inherited roles (bool)
        :return: List
        """
        rank = ROLE_RANKING[PROJECT_ROLE_DELEGATE]
        return self.get_roles(
            inherited=inherited,
            inherited_only=inherited_only,
            min_rank=rank,
            max_rank=rank,
        )

    def is_owner(self, user: 'SODARUser') -> bool:
        """
        Return True if user is owner in this project or inherits ownership from
        a parent category.

        :param user: SODARUser object
        :return: Boolean
        """
        if not user.is_authenticated:
            return False
        role_as = self.get_role(user)
        if role_as and role_as.role.rank == ROLE_RANKING[PROJECT_ROLE_OWNER]:
            return True
        return False

    def is_delegate(self, user: 'SODARUser') -> bool:
        """
        Return True if user is delegate in this project or inherits delegate
        status from a parent category.

        :param user: SODARUser object
        :return: Boolean
        """
        if not user.is_authenticated:
            return False
        role_as = self.get_role(user)
        if role_as and role_as.role.rank == ROLE_RANKING[PROJECT_ROLE_DELEGATE]:
            return True
        return False

    def is_owner_or_delegate(self, user: 'SODARUser') -> bool:
        """
        Return True if user is either an owner or a delegate in this project.
        Includes inherited assignments.

        :param user: SODARUser object
        :return: Boolean
        """
        if not user.is_authenticated:
            return False
        role_as = self.get_role(user)
        if role_as and role_as.role.rank in [
            ROLE_RANKING[PROJECT_ROLE_OWNER],
            ROLE_RANKING[PROJECT_ROLE_DELEGATE],
        ]:
            return True
        return False

    def get_members(self, inherited: bool = True) -> list['RoleAssignment']:
        """
        Return RoleAssignments for members of project excluding owner and
        delegates.

        :param inherited: Include inherited roles (boolean)
        :return: List of RoleAssignments
        """
        return self.get_roles(
            inherited=inherited,
            min_rank=Role.objects.get(name=PROJECT_ROLE_CONTRIBUTOR).rank,
        )

    def has_role(self, user: 'SODARUser', public: bool = True) -> bool:
        """
        Return whether user has roles in Project. Returns True if user has local
        role, inherits a role from a parent category, or if public access is
        enabled for the project. If public is set False, discount users with no
        actual role in a public access project.

        :param user: User object
        :param public: Boolean
        :return: Boolean
        """
        if (public and self.public_access) or self.get_role(user=user):
            return True
        return False

    def has_role_in_children(self, user: 'SODARUser') -> bool:
        """
        Return True if user has a role in any of the children in the project.
        Also returns true if public guest access is true for any child.

        :param user: User object
        :return: Boolean
        """
        if self.is_project():
            return False
        # User with role in self has at least the same role in children
        if self.has_role(user):
            return True
        children = self.get_children(flat=True)
        if (
            any([c.public_access is not None for c in children])
            or RoleAssignment.objects.filter(
                user=user, project__in=children
            ).count()
            > 0
        ):
            return True
        return False

    def get_source_site(self) -> Optional['RemoteSite']:
        """
        Return source site or None if this is a locally defined project.

        :return: RemoteProject object or None
        """
        if (
            settings.PROJECTROLES_SITE_MODE
            == SODAR_CONSTANTS['SITE_MODE_SOURCE']
        ):
            return None
        RemoteProject = apps.get_model(APP_NAME, 'RemoteProject')
        try:
            return RemoteProject.objects.get(
                project_uuid=self.sodar_uuid,
                site__mode=SODAR_CONSTANTS['SITE_MODE_SOURCE'],
            ).site
        except RemoteProject.DoesNotExist:
            pass
        return None

    def is_remote(self) -> bool:
        """
        Return True if current project has been retrieved from a remote SODAR
        site.

        :return: Boolean
        """
        if (
            settings.PROJECTROLES_SITE_MODE
            == SODAR_CONSTANTS['SITE_MODE_TARGET']
            and self.get_source_site()
        ):
            return True
        return False

    def is_revoked(self) -> bool:
        """
        Return True if remote access has been revoked for the project.

        :return: Boolean
        """
        if self.is_remote():
            remote_project = RemoteProject.objects.filter(
                project=self, site=self.get_source_site()
            ).first()
            if (
                remote_project
                and remote_project.level
                == SODAR_CONSTANTS['REMOTE_LEVEL_REVOKED']
            ):
                return True
        return False

    def set_public_access(self, role: Union['Role', str, None]):
        """
        Set project public_access value.

        If given as str, the Role object with its name coresponding to the
        value will be queried for an inserted.

        :param role: Role object, string or None
        """
        # NOTE: Validation no longer raises an exception (see #1404)
        if self.is_category() and role:
            raise ValidationError(CAT_PUBLIC_ACCESS_MSG)
        if isinstance(role, str):
            role = Role.objects.get(name=role)
        if self.public_access != role:
            self.public_access = role
            self.public_guest_access = True if role else False  # DEPRECATED
            self.save()
            self._update_public_children()  # Update for parents

    # TODO: Deprecated, remove in v1.3 (#1703)
    def set_public(self, public: bool = True):
        """
        Helper for setting value of public_access.

        DEPRECATED: Will be removed in v1.3. Use set_public_access() instead.

        :param public: Boolean (default=True)
        """
        logger.warning(SET_PUBLIC_DEPRECATE_MSG)
        # NOTE: Validation no longer raises an exception (see #1404)
        if self.is_category() and public:
            raise ValidationError(CAT_PUBLIC_ACCESS_MSG)
        role = Role.objects.get(name=PROJECT_ROLE_GUEST) if public else None
        self.public_access = role
        self.public_guest_access = public  # DEPRECATED
        self.save()
        self._update_public_children()  # Update for parents

    def set_archive(self, status: bool = True):
        """
        Helper for setting archive value. Raises ValidationError for categories.
        """
        if status != self.archive:
            self.archive = status
            self.save()

    def get_log_title(self, full_title: bool = False) -> str:
        """
        Return a logger-friendly title for the project.

        :param full_title: Display full title if True (boolean)
        :return: String
        """
        return '"{}" ({})'.format(
            self.full_title if full_title else self.title, self.sodar_uuid
        )


# Role -------------------------------------------------------------------------


def get_role_project_type_default() -> list[str]:
    """Return default value for Role.project_type"""
    return [PROJECT_TYPE_CATEGORY, PROJECT_TYPE_PROJECT]


class Role(models.Model):
    """Role definition, used to assign roles to projects in RoleAssignment"""

    #: Name of role
    name = models.CharField(
        max_length=64, unique=True, help_text='Name of role'
    )

    #: Role rank for determining role hierarchy
    rank = models.IntegerField(
        default=0,  # 0 = No rank
        help_text='Role rank for determining role hierarchy',
    )

    #: Allowed project types for the role
    project_types = ArrayField(
        models.CharField(max_length=64, blank=False),
        default=get_role_project_type_default,
        help_text='Allowed project types for the role',
    )

    #: Role description
    description = models.TextField(help_text='Role description')

    def __str__(self):
        return self.name

    def __repr__(self):
        return 'Role({})'.format(repr(self.name))


# RoleAssignment ---------------------------------------------------------------


class RoleAssignment(models.Model):
    """
    Assignment of an user to a role in a project. One local assignment per user
    is allowed for each project.

    One local project owner assignment is allowed for a project. Local project
    delegate assignements might be limited using PROJECTROLES_DELEGATE_LIMIT.

    Inherited role assignments can be accessed via the Project model API, e.g.
    Project.get_roles().
    """

    #: Project in which role is assigned
    project = models.ForeignKey(
        Project,
        related_name='local_roles',
        help_text='Project in which role is assigned',
        on_delete=models.CASCADE,
    )

    #: User for whom role is assigned
    user = models.ForeignKey(
        AUTH_USER_MODEL,
        related_name='roles',
        help_text='User for whom role is assigned',
        on_delete=models.CASCADE,
    )

    #: Role to be assigned
    role = models.ForeignKey(
        Role,
        related_name='assignments',
        help_text='Role to be assigned',
        on_delete=models.CASCADE,
    )

    #: RoleAssignment SODAR UUID
    sodar_uuid = models.UUIDField(
        default=uuid.uuid4, unique=True, help_text='RoleAssignment SODAR UUID'
    )

    class Meta:
        ordering = [
            'project__parent__title',
            'project__title',
            'role__name',
            'user__username',
        ]

    def __str__(self):
        return f'{self.project}: {self.role}: {self.user}'

    def __repr__(self):
        values = (self.project.title, self.user.username, self.role.name)
        return 'RoleAssignment({})'.format(', '.join(repr(v) for v in values))

    def _validate_project_type(self):
        """Validate type of project to ensure it is in allowed types"""
        if self.project.type not in self.role.project_types:
            raise ValidationError(
                ROLE_PROJECT_TYPE_ERROR_MSG.format(
                    project_type=self.project.type, role_name=self.role.name
                )
            )

    def _validate_user(self):
        """
        Validate fields to ensure user has only one role set for the project.
        """
        role_as = RoleAssignment.objects.filter(
            project=self.project, user=self.user
        ).first()
        if role_as and (not self.pk or role_as.pk != self.pk):
            raise ValidationError(
                f'Role {role_as.role} already set for {role_as.user} in '
                f'{role_as.project}'
            )

    def _validate_owner(self):
        """
        Validate role to ensure no more than one project owner is assigned to a
        project.
        """
        if self.role.name == PROJECT_ROLE_OWNER:
            owner = self.project.get_owner()
            if owner and (not self.pk or owner.pk != self.pk):
                raise ValidationError(
                    f'User {owner} already set as owner of {self.project}'
                )

    def _validate_delegate(self):
        """
        Validate role to ensure no more than the allowed amount of project
        delegates are assigned to a project.
        """
        if (
            self.role.name != PROJECT_ROLE_DELEGATE
            or self.project.is_remote()  # No validation for remote projects
        ):
            return
        del_limit = getattr(settings, 'PROJECTROLES_DELEGATE_LIMIT', 1)
        delegates = self.project.get_delegates(inherited=False)
        if 0 < del_limit <= len(delegates) and (
            not self.pk
            or self.project.local_roles.filter(
                role__name=PROJECT_ROLE_DELEGATE, pk=self.pk
            )
            is None
        ):
            raise ValidationError(
                f'The local delegate limit for this project ({del_limit}) has '
                f'already been reached.'
            )

    def save(self, *args, **kwargs):
        """Version of save() to include custom validation for RoleAssignment"""
        self._validate_project_type()
        self._validate_user()
        self._validate_owner()
        self._validate_delegate()
        super().save(*args, **kwargs)


# AppSetting -------------------------------------------------------------------


class AppSettingManager(models.Manager):
    """Manager for custom table-level AppSetting queries"""

    def get_setting_value(
        self,
        plugin_name: str,
        setting_name: str,
        project: Optional[Project] = None,
        user: Optional['SODARUser'] = None,
    ) -> Any:
        """
        Return value of setting_name for plugin_name in project or for user.

        Note that project and/or user must be set.

        :param plugin_name: App plugin name (string)
        :param setting_name: Name of setting (string)
        :param project: Project object or pk
        :param user: User object or pk
        :return: Value (string, int, bool, dict, list or None)
        :raise: AppSetting.DoesNotExist if setting is not found
        """
        query_parameters = {
            'name': setting_name,
            'project': project,
            'user': user,
        }
        if not plugin_name == APP_NAME:
            query_parameters['app_plugin__name'] = plugin_name
        setting = super().get_queryset().get(**query_parameters)
        return setting.get_value()


class AppSetting(models.Model):
    """
    Project and users settings value.

    The settings are defined in the "app_settings" member in a SODAR project
    app's plugin. The scope of each setting can be either "USER" or "PROJECT",
    defined for each setting in app_settings. Project AND user-specific settings
    or settings which don't belong to either are are currently not supported.
    """

    #: App to which the setting belongs
    app_plugin = models.ForeignKey(
        Plugin,
        null=True,
        unique=False,
        related_name='settings',
        help_text='App to which the setting belongs',
        on_delete=models.CASCADE,
    )

    #: Project to which the setting belongs
    project = models.ForeignKey(
        Project,
        null=True,
        blank=True,
        related_name='settings',
        help_text='Project to which the setting belongs',
        on_delete=models.CASCADE,
    )

    #: Project to which the setting belongs
    user = models.ForeignKey(
        AUTH_USER_MODEL,
        null=True,
        blank=True,
        related_name='user_settings',
        help_text='User to which the setting belongs',
        on_delete=models.CASCADE,
    )

    #: Name of the setting
    name = models.CharField(
        max_length=255, unique=False, help_text='Name of the setting'
    )

    #: Type of the setting
    type = models.CharField(
        max_length=64, unique=False, help_text='Type of the setting'
    )

    #: Value of the setting
    value = models.CharField(
        unique=False,
        null=True,
        blank=True,
        help_text='Value of the setting',
    )

    #: Optional JSON value for the setting
    value_json = models.JSONField(
        null=True, default=dict, help_text='Optional JSON value for the setting'
    )

    #: Setting visibility in forms
    user_modifiable = models.BooleanField(
        default=True, help_text='Setting visibility in forms'
    )

    #: AppSetting SODAR UUID
    sodar_uuid = models.UUIDField(
        default=uuid.uuid4, unique=True, help_text='AppSetting SODAR UUID'
    )

    # Set manager for custom queries
    objects = AppSettingManager()

    class Meta:
        ordering = ['project__title', 'app_plugin__name', 'name']
        unique_together = ['project', 'user', 'app_plugin', 'name']

    def __str__(self):
        plugin_name = self.app_plugin.name if self.app_plugin else APP_NAME
        if self.project:
            label = self.project.title
        elif self.user:
            label = self.user.username
        else:
            label = 'SITE'
        return f'{label}: {plugin_name} / {self.name}'

    def __repr__(self):
        values = (
            self.project.title if self.project else None,
            self.user.username if self.user else None,
            self.app_plugin.name if self.app_plugin else APP_NAME,
            self.name,
        )
        return 'AppSetting({})'.format(', '.join(repr(v) for v in values))

    def save(self, *args, **kwargs):
        """Version of save() to convert 'value' data according to 'type'"""
        if self.type == APP_SETTING_TYPE_BOOLEAN:
            self.value = str(int(self.value))
        elif self.type == APP_SETTING_TYPE_INTEGER:
            self.value = str(self.value)
        super().save(*args, **kwargs)

    # Custom row-level functions

    def get_value(self) -> Any:
        """Return value of the setting in the format specified in 'type'"""
        if self.type == APP_SETTING_TYPE_INTEGER:
            return int(self.value)
        elif self.type == APP_SETTING_TYPE_BOOLEAN:
            return bool(int(self.value))
        elif self.type == APP_SETTING_TYPE_JSON:
            return self.value_json
        return self.value


# ProjectInvite ----------------------------------------------------------------


class ProjectInvite(models.Model):
    """
    Invite which is sent to a non-logged in user, determining their role in
    the project.
    """

    #: Email address of the person to be invited
    email = models.EmailField(
        unique=False,
        null=False,
        blank=False,
        help_text='Email address of the person to be invited',
    )

    #: Project to which the person is invited
    project = models.ForeignKey(
        Project,
        null=False,
        related_name='invites',
        help_text='Project to which the person is invited',
        on_delete=models.CASCADE,
    )

    #: Role assigned to the person
    role = models.ForeignKey(
        Role,
        null=False,
        help_text='Role assigned to the person',
        on_delete=models.CASCADE,
    )

    #: User who issued the invite
    issuer = models.ForeignKey(
        AUTH_USER_MODEL,
        null=False,
        related_name='issued_invites',
        help_text='User who issued the invite',
        on_delete=models.CASCADE,
    )

    #: DateTime of invite creation
    date_created = models.DateTimeField(
        auto_now_add=True, help_text='DateTime of invite creation'
    )

    #: Expiration of invite as DateTime
    date_expire = models.DateTimeField(
        null=False, help_text='Expiration of invite as DateTime'
    )

    #: Message to be included in the invite email (optional)
    message = models.TextField(
        blank=True,
        help_text='Message to be included in the invite email (optional)',
    )

    #: Secret token provided to user with the invite
    secret = models.CharField(
        max_length=255,
        unique=True,
        blank=False,
        null=False,
        help_text='Secret token provided to user with the invite',
    )

    #: Status of the invite (False if claimed or revoked)
    active = models.BooleanField(
        default=True,
        help_text='Status of the invite (False if claimed or revoked)',
    )

    #: ProjectInvite SODAR UUID
    sodar_uuid = models.UUIDField(
        default=uuid.uuid4, unique=True, help_text='ProjectInvite SODAR UUID'
    )

    class Meta:
        ordering = ['project__title', 'email', 'role__name']

    def __str__(self):
        return '{}: {} ({}){}'.format(
            self.project,
            self.email,
            self.role.name,
            ' [ACTIVE]' if self.active else '',
        )

    def __repr__(self):
        values = (self.project.title, self.email, self.role.name, self.active)
        return 'ProjectInvite({})'.format(', '.join(repr(v) for v in values))

    @classmethod
    def _get_date_expire(cls) -> datetime:
        """
        Return expiry date based on current date + INVITE_EXPIRY_DAYS

        :return: DateTime object
        """
        return timezone.now() + timezone.timedelta(
            days=settings.PROJECTROLES_INVITE_EXPIRY_DAYS
        )

    def save(self, *args, **kwargs):
        if not self.pk and not self.date_expire:  # Set date_expire on create
            self.date_expire = self._get_date_expire()
        super().save(*args, **kwargs)

    # Custom row-level functions

    def is_ldap(self) -> bool:
        """
        Return True if invite is intended for an LDAP user.

        :return: Boolean
        """
        # Only consider LDAP if enabled in Django settings
        if not settings.ENABLE_LDAP:
            return False
        # Check if domain is associated with LDAP
        domain = self.email.split('@')[1].lower()
        domain_no_tld = domain.split('.')[0].lower()
        ldap_domains = [
            getattr(settings, 'AUTH_LDAP_USERNAME_DOMAIN', '').lower(),
            getattr(settings, 'AUTH_LDAP2_USERNAME_DOMAIN', '').lower(),
        ]
        alt_domains = [
            a.lower() for a in getattr(settings, 'LDAP_ALT_DOMAINS', [])
        ]
        if domain_no_tld in ldap_domains or domain in alt_domains:
            return True
        return False

    def get_url(self, request: HttpRequest) -> str:
        """
        Return invite URL for a project invitation.

        :param request: HttpRequest object
        :return: URL (string)
        """
        return request.build_absolute_uri(
            reverse(
                'projectroles:invite_accept', kwargs={'secret': self.secret}
            )
        )

    def reset_date_expire(self):
        """
        Reset date_expire to current date plus defined expiry days. Saves the
        object.
        """
        self.date_expire = self._get_date_expire()
        self.save()


# RemoteSite -------------------------------------------------------------------


class RemoteSite(models.Model):
    """Remote SODAR site"""

    #: Site name
    name = models.CharField(
        max_length=255,
        unique=True,
        blank=False,
        null=False,
        help_text='Site name',
    )

    #: Site URL
    url = models.URLField(
        max_length=2000,
        blank=False,
        null=False,
        unique=False,
        help_text='Site URL',
    )

    #: Site mode
    mode = models.CharField(
        max_length=64,
        unique=False,
        blank=False,
        null=False,
        default=SODAR_CONSTANTS['SITE_MODE_TARGET'],
        help_text='Site mode',
    )

    #: Site description
    description = models.TextField(help_text='Site description')

    #: Secret token used to connect to the master site
    secret = models.CharField(
        max_length=255,
        unique=False,
        blank=False,
        null=True,  # Can be NULL for Peer Mode
        help_text='Secret token for connecting to the source site',
    )

    #: RemoteSite visibilty to users
    user_display = models.BooleanField(
        default=True, unique=False, help_text='Display site to users'
    )

    #: RemoteSite project access modifiability for owners and delegates
    owner_modifiable = models.BooleanField(
        default=True,
        unique=False,
        help_text='Allow owners and delegates to modify project access for '
        'this site',
    )

    #: RemoteSite relation UUID (local)
    sodar_uuid = models.UUIDField(
        default=uuid.uuid4,
        unique=True,
        help_text='RemoteSite relation UUID (local)',
    )

    class Meta:
        ordering = ['name']
        unique_together = ['url', 'mode', 'secret']

    def __str__(self):
        return f'{self.name} ({self.mode})'

    def __repr__(self):
        values = (self.name, self.mode, self.url)
        return 'RemoteSite({})'.format(', '.join(repr(v) for v in values))

    def _validate_mode(self):
        """Validate mode value"""
        if self.mode not in SODAR_CONSTANTS['SITE_MODES']:
            raise ValidationError(f'Mode "{self.mode}" not found in SITE_MODES')

    def save(self, *args, **kwargs):
        """Version of save() to include custom validation"""
        self._validate_mode()
        super().save(*args, **kwargs)

    # Custom row-level functions

    def get_access_date(self) -> Optional[datetime]:
        """Return date of latest project access by remote site or None"""
        projects = (
            RemoteProject.objects.filter(site=self)
            .exclude(date_access__isnull=True)
            .order_by('-date_access')
        )
        if projects.count() > 0:
            return projects.first().date_access
        return None

    def get_url(self) -> str:
        """Return sanitized site URL"""
        if self.url[-1] == '/':
            return self.url[:-1]
        return self.url


# RemoteProject ----------------------------------------------------------------


class RemoteProject(models.Model):
    """Remote project relation"""

    #: Related project UUID
    project_uuid = models.UUIDField(
        default=None, unique=False, help_text='Project UUID'
    )

    #: Related project object (if created locally)
    project = models.ForeignKey(
        Project,
        related_name='remotes',
        blank=True,
        null=True,
        help_text='Related project object (if created locally)',
        on_delete=models.CASCADE,
    )

    #: Related remote SODAR site
    site = models.ForeignKey(
        RemoteSite,
        null=False,
        related_name='projects',
        help_text='Remote SODAR site',
        on_delete=models.CASCADE,
    )

    #: Project access level
    level = models.CharField(
        max_length=255,
        unique=False,
        blank=False,
        null=False,
        default=SODAR_CONSTANTS['REMOTE_LEVEL_NONE'],
        help_text='Project access level',
    )

    #: DateTime of last access from/to remote site
    date_access = models.DateTimeField(
        null=True,
        auto_now_add=False,
        help_text='DateTime of last access from/to remote site',
    )

    #: RemoteProject relation UUID (local)
    sodar_uuid = models.UUIDField(
        default=uuid.uuid4,
        unique=True,
        help_text='RemoteProject relation UUID (local)',
    )

    class Meta:
        ordering = ['site__name', 'project_uuid']

    def save(self, *args, **kwargs):
        # NOTE: Can't use unique constraint with foreign key relation
        rp = self.__class__.objects.filter(
            project_uuid=self.project_uuid, site=self.site
        ).first()
        if rp and rp.id != self.id:
            raise ValidationError(REMOTE_PROJECT_UNIQUE_MSG)
        super().save(*args, **kwargs)

    def __str__(self):
        return f'{self.site.name}: {self.project_uuid} ({self.site.mode})'

    def __repr__(self):
        values = (self.site.name, str(self.project_uuid), self.site.mode)
        return 'RemoteProject({})'.format(', '.join(repr(v) for v in values))

    # Custom row-level functions

    def get_project(self) -> Project:
        """Get the related Project object"""
        return (
            self.project
            or Project.objects.filter(sodar_uuid=self.project_uuid).first()
        )


# User Models ------------------------------------------------------------------


class SODARUser(AbstractUser):
    """
    SODAR compatible abstract user model. Use this on your SODAR Core based
    site.
    """

    # First Name and Last Name do not cover name patterns
    # around the globe.
    name = models.CharField(_('Name of User'), blank=True, max_length=255)

    #: User SODAR UUID
    sodar_uuid = models.UUIDField(
        default=uuid.uuid4, unique=True, help_text='User SODAR UUID'
    )

    class Meta:
        abstract = True
        ordering = ['name', 'username']

    def __str__(self):
        return self.username

    def save(self, *args, **kwargs):
        super().save(*args, **kwargs)
        self.set_group()  # Called here to ensure we always have group

    def get_full_name(self) -> str:
        """Return full name or username if not set"""
        if hasattr(self, 'name') and self.name:
            return self.name
        elif self.first_name and self.last_name:
            return f'{self.first_name} {self.last_name}'
        return self.username

    def get_display_name(self, inc_user: bool = False) -> str:
        """
        Return user name for displaying in UI.

        :param inc_user: Include username if true (boolean, default=False)
        :return: String
        """
        ret = self.get_full_name()
        if ret != self.username and inc_user:
            ret += f' ({self.username})'
        return ret

    def get_form_label(self, email: bool = False) -> str:
        """
        Return user label with full name, username and optional email.

        :param email: Return email if True (boolean, default=False)
        :return: String
        """
        ret = self.get_full_name()
        if ret != self.username:
            ret += f' ({self.username})'
        if email and self.email:
            ret += f' <{self.email}>'
        return ret

    def get_auth_type(self) -> str:
        """
        Return user authentication type: OIDC, LDAP or local.

        :return: String which may equal AUTH_TYPE_OIDC, AUTH_TYPE_LDAP or
                 AUTH_TYPE_LOCAL.
        """
        groups = [g.name for g in self.groups.all()]
        if 'oidc' in groups:
            return AUTH_TYPE_OIDC
        elif (
            self.username.find('@') != -1
            and self.username.split('@')[1].lower() in groups
        ):
            return AUTH_TYPE_LDAP
        return AUTH_TYPE_LOCAL

    def is_local(self) -> bool:
        """
        Return True if user is of type AUTH_TYPE_LOCAL.

        :return: Boolean
        """
        return self.get_auth_type() == AUTH_TYPE_LOCAL

    def set_group(self) -> Optional[str]:
        """
        Set user group based on user name or social auth provider.

        :return: Group name if set (string or None)
        """
        social_auth = getattr(self, 'social_auth', None)
        if social_auth:
            social_auth = social_auth.first()
        ldap_domains = [
            getattr(settings, 'AUTH_LDAP_USERNAME_DOMAIN', '').upper(),
            getattr(settings, 'AUTH_LDAP2_USERNAME_DOMAIN', '').upper(),
        ]
        # OIDC user group
        if social_auth and social_auth.provider == AUTH_PROVIDER_OIDC:
            group_name = AUTH_PROVIDER_OIDC
        # LDAP domain user groups
        elif (
            self.username.find('@') != -1
            and self.username.split('@')[1].upper() in ldap_domains
        ):
            group_name = self.username.split('@')[1].lower()
        # System user group for local users
        else:
            group_name = SYSTEM_USER_GROUP
        group, created = Group.objects.get_or_create(name=group_name)
        if group not in self.groups.all():
            group.user_set.add(self)
            # Remove system group if previously added
            if (
                group_name != SYSTEM_USER_GROUP
                and self.groups.filter(name=SYSTEM_USER_GROUP).count() > 0
            ):
                system_group = Group.objects.get(name=SYSTEM_USER_GROUP)
                system_group.user_set.remove(self)
                logger.debug(
                    f'Removed system user group "{SYSTEM_USER_GROUP}" from '
                    f'{self.username}'
                )
            logger.info(f'Set user group "{group_name}" for {self.username}')
            return group_name
        return None

    def update_full_name(self) -> str:
        """
        Update full name of user.

        :return: User full name (string)
        """
        # Save user name from first_name and last_name into name
        full_name = ''
        if self.first_name != '':
            full_name = self.first_name + (
                ' ' + self.last_name if self.last_name != '' else ''
            )
        if self.name != full_name:
            self.name = full_name
            self.save()
            logger.info(
                f'Full name updated for user {self.username}: {self.name}'
            )
        return self.name

    def update_ldap_username(self) -> str:
        """
        Update username for an LDAP user.

        :return: User username (string)
        """
        # Make domain in username uppercase
        if (
            self.username.find('@') != -1
            and self.username.split('@')[1].islower()
        ):
            u_split = self.username.split('@')
            self.username = u_split[0] + '@' + u_split[1].upper()
            self.save()
        return self.username


class SODARUserAdditionalEmail(models.Model):
    """
    Model representing an additional email address for a user. Stores
    information for email verification.
    """

    #: User for whom the email is assigned
    user = models.ForeignKey(
        AUTH_USER_MODEL,
        related_name='additional_emails',
        help_text='User for whom the email is assigned',
        on_delete=models.CASCADE,
    )

    #: Email address
    email = models.EmailField(
        unique=False,
        null=False,
        blank=False,
        help_text='Email address',
    )

    #: Email verification status
    verified = models.BooleanField(
        default=False, help_text='Email verification status'
    )

    #: Secret token for email verification
    secret = models.CharField(
        max_length=255,
        unique=True,
        blank=False,
        null=True,
        help_text='Secret token for email verification',
    )

    #: DateTime of creation
    date_created = models.DateTimeField(
        auto_now_add=True, help_text='DateTime of creation'
    )

    #: DateTime of last modification
    date_modified = models.DateTimeField(
        auto_now=True, help_text='DateTime of last modification'
    )

    #: SODARUserAdditionalEmail SODAR UUID
    sodar_uuid = models.UUIDField(
        default=uuid.uuid4,
        unique=True,
        help_text='SODARUserAdditionalEmail SODAR UUID',
    )

    class Meta:
        ordering = ['user__username', 'email']
        unique_together = ['user', 'email']

    def __str__(self):
        return f'{self.user.username}: {self.email}'

    def __repr__(self):
        values = (
            self.user.username,
            self.email,
            str(self.verified),
            self.secret,
            str(self.sodar_uuid),
        )
        return 'SODARUserAdditionalEmail({})'.format(
            ', '.join(repr(v) for v in values)
        )

    def _validate_email_unique(self):
        """
        Assert the same email has not yet been set for the user.
        """
        if self.email == self.user.email:
            raise ValidationError(
                ADD_EMAIL_ALREADY_SET_MSG.format(email_type='primary')
            )

    def save(self, *args, **kwargs):
        self._validate_email_unique()
        super().save(*args, **kwargs)



"""UI tests for the timeline app"""

import json

from django.urls import reverse
from django.utils.timezone import localtime
from django.test import override_settings
from urllib.parse import urlencode

from selenium.webdriver.common.by import By
from selenium.webdriver.support.wait import WebDriverWait
from selenium.webdriver.support import expected_conditions as EC

# Projectroles dependency
from projectroles.models import SODAR_CONSTANTS
from projectroles.plugins import PluginAPI
from projectroles.tests.test_ui import UITestBase

from timeline.models import TL_STATUS_OK
from timeline.tests.test_models import (
    TimelineEventMixin,
    TimelineEventStatusMixin,
    EXTRA_DATA,
)


plugin_api = PluginAPI()


# SODAR constants
PROJECT_ROLE_OWNER = SODAR_CONSTANTS['PROJECT_ROLE_OWNER']
PROJECT_ROLE_DELEGATE = SODAR_CONSTANTS['PROJECT_ROLE_DELEGATE']
PROJECT_ROLE_CONTRIBUTOR = SODAR_CONSTANTS['PROJECT_ROLE_CONTRIBUTOR']
PROJECT_ROLE_GUEST = SODAR_CONSTANTS['PROJECT_ROLE_GUEST']
PROJECT_TYPE_CATEGORY = SODAR_CONSTANTS['PROJECT_TYPE_CATEGORY']
PROJECT_TYPE_PROJECT = SODAR_CONSTANTS['PROJECT_TYPE_PROJECT']

# Local constants
APP_NAME_PR = 'projectroles'


class TestProjectListView(
    TimelineEventMixin, TimelineEventStatusMixin, UITestBase
):
    """Tests for the timeline project list view UI"""

    def setUp(self):
        super().setUp()
        self.timeline = plugin_api.get_backend_api('timeline_backend')
        # Init default event
        self.event = self.timeline.add_event(
            project=self.project,
            app_name=APP_NAME_PR,
            user=self.superuser,
            event_name='test_event',
            description='description',
            extra_data=EXTRA_DATA,
            status_type=self.timeline.TL_STATUS_OK,
        )
        # Init classified event
        self.classified_event = self.timeline.add_event(
            project=self.project,
            app_name=APP_NAME_PR,
            user=self.superuser,
            event_name='classified_event',
            description='description',
            extra_data=EXTRA_DATA,
            classified=True,
        )

    def test_render(self):
        """Test visibility of events in project event list"""
        expected = [
            (self.superuser, 2),
            (self.user_owner_cat, 2),
            (self.user_delegate_cat, 2),
            (self.user_contributor_cat, 1),
            (self.user_guest_cat, 1),
            (self.user_owner, 2),
            (self.user_delegate, 2),
            (self.user_contributor, 1),
            (self.user_guest, 1),
        ]
        url = reverse(
            'timeline:list_project', kwargs={'project': self.project.sodar_uuid}
        )
        self.assert_element_count(expected, url, 'sodar-tl-list-event')

    def test_render_no_user(self):
        """Test rendering with an event without user"""
        self.event.user = None
        expected = [
            (self.superuser, 2),
            (self.user_owner_cat, 2),
            (self.user_delegate_cat, 2),
            (self.user_contributor_cat, 1),
            (self.user_guest_cat, 1),
            (self.user_owner, 2),
            (self.user_delegate, 2),
            (self.user_contributor, 1),
            (self.user_guest, 1),
        ]
        url = reverse(
            'timeline:list_project', kwargs={'project': self.project.sodar_uuid}
        )
        self.assert_element_count(expected, url, 'sodar-tl-list-event')

    def test_render_object(self):
        """Test visibility of object related events in project list"""
        # Add user as an object reference
        self.obj_ref = self.event.add_object(
            obj=self.superuser, label='user', name=self.superuser.username
        )
        self.classified_obj_ref = self.classified_event.add_object(
            obj=self.superuser, label='user', name=self.superuser.username
        )
        expected = [
            (self.superuser, 2),
            (self.user_owner_cat, 2),
            (self.user_delegate_cat, 2),
            (self.user_contributor_cat, 1),
            (self.user_guest_cat, 1),
            (self.user_owner, 2),
            (self.user_delegate, 2),
            (self.user_contributor, 1),
            (self.user_guest, 1),
        ]
        url = reverse(
            'timeline:list_object',
            kwargs={
                'project': self.project.sodar_uuid,
                'object_model': self.obj_ref.object_model,
                'object_uuid': self.obj_ref.object_uuid,
            },
        )
        self.assert_element_count(expected, url, 'sodar-tl-list-event')

    def test_render_details(self):
        """Test visibility of events on the project details page"""
        expected = [
            (self.superuser, 2),
            (self.user_owner_cat, 2),
            (self.user_delegate_cat, 2),
            (self.user_contributor_cat, 1),
            (self.user_guest_cat, 1),
            (self.user_owner, 2),
            (self.user_delegate, 2),
            (self.user_contributor, 1),
            (self.user_guest, 1),
        ]
        url = reverse(
            'projectroles:detail', kwargs={'project': self.project.sodar_uuid}
        )
        self.assert_element_count(expected, url, 'sodar-tl-list-event')


class TestSiteListView(
    TimelineEventMixin, TimelineEventStatusMixin, UITestBase
):
    """Tests for the timeline site-wide list view UI"""

    def setUp(self):
        super().setUp()
        self.timeline = plugin_api.get_backend_api('timeline_backend')
        # Init default event
        self.event = self.timeline.add_event(
            project=None,
            app_name=APP_NAME_PR,
            user=self.superuser,
            event_name='test_event',
            description='description',
            extra_data=EXTRA_DATA,
            status_type=self.timeline.TL_STATUS_OK,
        )
        # Init classified event
        self.classified_event = self.timeline.add_event(
            project=None,
            app_name=APP_NAME_PR,
            user=self.superuser,
            event_name='classified_event',
            description='description',
            extra_data=EXTRA_DATA,
            classified=True,
        )

    def test_render(self):
        """Test visibility of events in the site-wide event list"""
        expected = [
            (self.superuser, 2),
            (self.user_owner_cat, 1),
            (self.user_delegate_cat, 1),
            (self.user_contributor_cat, 1),
            (self.user_guest_cat, 1),
            (self.user_owner, 1),
            (self.user_delegate, 1),
            (self.user_contributor, 1),
            (self.user_guest, 1),
        ]
        url = reverse('timeline:list_site')
        self.assert_element_count(expected, url, 'sodar-tl-list-event')

    def test_render_object(self):
        """Test visibility of object related events in site-wide event list"""
        obj_ref = self.event.add_object(
            obj=self.superuser, label='user', name=self.superuser.username
        )
        self.classified_event.add_object(
            obj=self.superuser, label='user', name=self.superuser.username
        )
        expected = [
            (self.superuser, 2),
            (self.user_owner_cat, 1),
            (self.user_delegate_cat, 1),
            (self.user_contributor_cat, 1),
            (self.user_guest_cat, 1),
            (self.user_owner, 1),
            (self.user_delegate, 1),
            (self.user_contributor, 1),
            (self.user_guest, 1),
        ]
        url = reverse(
            'timeline:list_object_site',
            kwargs={
                'object_model': obj_ref.object_model,
                'object_uuid': obj_ref.object_uuid,
            },
        )
        self.assert_element_count(expected, url, 'sodar-tl-list-event')

    def test_render_back_button(self):
        """Test visibility of back button in object view"""
        obj_ref = self.event.add_object(
            obj=self.superuser, label='user', name=self.superuser.username
        )
        url = reverse(
            'timeline:list_object_site',
            kwargs={
                'object_model': obj_ref.object_model,
                'object_uuid': obj_ref.object_uuid,
            },
        )
        self.login_and_redirect(
            self.superuser, url, wait_elem=None, wait_loc='ID'
        )
        self.assertIsNotNone(
            self.selenium.find_element(By.ID, 'sodar-tl-back-btn')
        )


class TestAdminTimelineView(
    TimelineEventMixin, TimelineEventStatusMixin, UITestBase
):
    """Tests for AdminTimelineView"""

    def setUp(self):
        super().setUp()
        self.timeline = plugin_api.get_backend_api('timeline_backend')
        self.url = reverse('timeline:list_admin')
        # Init default event
        self.event = self.timeline.add_event(
            project=self.project,
            app_name=APP_NAME_PR,
            user=self.superuser,
            event_name='test_event',
            description='description',
            extra_data=EXTRA_DATA,
            status_type=self.timeline.TL_STATUS_OK,
        )
        # Init default site event
        self.site_event = self.timeline.add_event(
            project=None,
            app_name=APP_NAME_PR,
            user=self.superuser,
            event_name='test_site_event',
            description='description',
            extra_data=EXTRA_DATA,
            status_type=self.timeline.TL_STATUS_OK,
        )
        # Init classified event
        self.classified_event = self.timeline.add_event(
            project=None,
            app_name=APP_NAME_PR,
            user=self.superuser,
            event_name='classified_event',
            description='description',
            extra_data=EXTRA_DATA,
            classified=True,
        )

    def test_render(self):
        """Test visibility of events in the admin event list"""
        expected = [(self.superuser, 3)]
        self.assert_element_count(expected, self.url, 'sodar-tl-list-event')

    def test_badge(self):
        """Test visibility of badges in description of event"""
        self.login_and_redirect(
            self.superuser, self.url, wait_elem=None, wait_loc='ID'
        )
        self.assertIsNotNone(self.selenium.find_element(By.CLASS_NAME, 'badge'))


class TestModals(TimelineEventMixin, TimelineEventStatusMixin, UITestBase):
    """Test UI of modals in timeline event list"""

    def setUp(self):
        super().setUp()
        self.timeline = plugin_api.get_backend_api('timeline_backend')
        # Init default event
        self.event = self.timeline.add_event(
            project=self.project,
            app_name=APP_NAME_PR,
            user=self.superuser,
            event_name='test_event',
            description='description',
            extra_data=json.dumps(EXTRA_DATA),
            status_type=self.timeline.TL_STATUS_OK,
        )
        self.url = reverse(
            'timeline:list_project', kwargs={'project': self.project.sodar_uuid}
        )

    def test_details_modal(self):
        """Test details modal"""
        self.login_and_redirect(
            self.superuser, self.url, wait_elem=None, wait_loc='ID'
        )
        link = self.selenium.find_element(By.CLASS_NAME, 'sodar-tl-link-detail')
        link.click()
        WebDriverWait(self.selenium, 10).until(
            EC.presence_of_element_located((By.ID, 'sodar-tl-table-detail'))
        )
        self.assertIsNotNone(
            self.selenium.find_element(By.CLASS_NAME, 'sodar-card-table')
        )
        title = self.selenium.find_element(By.CLASS_NAME, 'modal-title')
        self.assertIn('Event Details: ', title.text)
        elem = self.selenium.find_element(
            By.CLASS_NAME, 'sodar-tl-event-detail-timestamp'
        )
        self.assertEqual(
            elem.text,
            localtime(self.event.get_timestamp()).strftime('%Y-%m-%d %H:%M:%S'),
        )
        elem = self.selenium.find_element(
            By.CLASS_NAME, 'sodar-tl-event-detail-description'
        )
        self.assertEqual(elem.text, self.event.get_status().description)
        elem = self.selenium.find_element(
            By.CLASS_NAME, 'sodar-tl-event-detail-status-type'
        )
        self.assertEqual(elem.text, TL_STATUS_OK)

    def test_details_modal_description_line_break(self):
        """Test details modal with line breaks in description"""
        desc = 'Test\nwith\nline breaks'
        status = self.event.get_status()
        status.description = desc
        status.save()
        self.login_and_redirect(
            self.superuser, self.url, wait_elem=None, wait_loc='ID'
        )
        link = self.selenium.find_element(By.CLASS_NAME, 'sodar-tl-link-detail')
        link.click()
        WebDriverWait(self.selenium, 10).until(
            EC.presence_of_element_located((By.ID, 'sodar-tl-table-detail'))
        )
        elem = self.selenium.find_element(
            By.CLASS_NAME, 'sodar-tl-event-detail-description'
        )
        self.assertEqual(
            elem.get_attribute('innerHTML'), desc.replace('\n', '<br>')
        )

    def test_extra_data_modal(self):
        """Test extra data modal"""
        self.login_and_redirect(
            self.superuser, self.url, wait_elem=None, wait_loc='ID'
        )
        link = self.selenium.find_element(
            By.CLASS_NAME, 'sodar-tl-link-extra-data'
        )
        link.click()
        WebDriverWait(self.selenium, 10).until(
            EC.presence_of_element_located((By.CLASS_NAME, 'sodar-tl-json'))
        )
        title = self.selenium.find_element(By.CLASS_NAME, 'modal-title')
        self.assertIn('Extra', title.text)
        body = self.selenium.find_element(By.CLASS_NAME, 'modal-body')
        self.assertIn('"{"test_key": "test_val"}"', body.text)

    def test_status_extra_data_modal(self):
        """Test status extra data modal"""
        status = self.event.get_status()
        status.extra_data = {'test': 170}
        status.save()
        self.login_and_redirect(
            self.superuser, self.url, wait_elem=None, wait_loc='ID'
        )
        link = self.selenium.find_element(By.CLASS_NAME, 'sodar-tl-link-detail')
        link.click()
        WebDriverWait(self.selenium, 10).until(
            EC.presence_of_element_located(
                (By.CLASS_NAME, 'sodar-tl-link-status-extra-data')
            )
        )
        self.assertIsNotNone(
            self.selenium.find_element(
                By.CLASS_NAME, 'sodar-tl-link-status-extra-data'
            )
        )

    def test_copy_button(self):
        """Test copy button"""
        self.login_and_redirect(
            self.superuser, self.url, wait_elem=None, wait_loc='ID'
        )
        link = self.selenium.find_element(
            By.CLASS_NAME, 'sodar-tl-link-extra-data'
        )
        link.click()
        WebDriverWait(self.selenium, 10).until(
            EC.presence_of_element_located((By.CLASS_NAME, 'sodar-tl-copy-btn'))
        )
        button = self.selenium.find_element(By.CLASS_NAME, 'sodar-tl-copy-btn')
        button.click()


class TestSearch(TimelineEventMixin, TimelineEventStatusMixin, UITestBase):
    """Tests for the project search UI functionalities"""

    def setUp(self):
        super().setUp()
        self.timeline = plugin_api.get_backend_api('timeline_backend')
        # Init default event
        self.event = self.timeline.add_event(
            project=self.project,
            app_name=APP_NAME_PR,
            user=self.superuser,
            event_name='test_event',
            description='description',
            extra_data=EXTRA_DATA,
            status_type=self.timeline.TL_STATUS_OK,
        )
        self.make_event_status(
            event=self.event,
            status_type=self.timeline.TL_STATUS_SUBMIT,
            description=self.timeline.TL_STATUS_SUBMIT,
            extra_data=EXTRA_DATA,
        )
        # Init default site event
        self.site_event = self.timeline.add_event(
            project=None,
            app_name=APP_NAME_PR,
            user=self.superuser,
            event_name='test_site_event',
            description='description',
            extra_data=EXTRA_DATA,
            status_type=self.timeline.TL_STATUS_OK,
        )
        self.make_event_status(
            event=self.site_event,
            status_type=self.timeline.TL_STATUS_SUBMIT,
            description=self.timeline.TL_STATUS_SUBMIT,
            extra_data=EXTRA_DATA,
        )
        # Init classified event
        self.classified_event = self.timeline.add_event(
            project=None,
            app_name=APP_NAME_PR,
            user=self.superuser,
            event_name='classified_event',
            description='description',
            extra_data=EXTRA_DATA,
            classified=True,
        )
        self.make_event_status(
            event=self.classified_event,
            status_type=self.timeline.TL_STATUS_SUBMIT,
            description=self.timeline.TL_STATUS_SUBMIT,
            extra_data=EXTRA_DATA,
        )
        self.classified_site_event = self.timeline.add_event(
            project=None,
            app_name=APP_NAME_PR,
            user=self.superuser,
            event_name='classified_site_event',
            description='description',
            extra_data=EXTRA_DATA,
            classified=True,
        )
        self.make_event_status(
            event=self.classified_site_event,
            status_type=self.timeline.TL_STATUS_SUBMIT,
            description=self.timeline.TL_STATUS_SUBMIT,
            extra_data=EXTRA_DATA,
        )

    def test_search_results(self):
        """Test search results"""
        expected = [
            (self.superuser, 4),
            (self.user_owner_cat, 2),
            (self.user_delegate_cat, 2),
            (self.user_contributor_cat, 2),
            (self.user_guest_cat, 2),
            (self.user_viewer_cat, 1),
            (self.user_finder_cat, 1),
            (self.user_owner, 2),
            (self.user_delegate, 2),
            (self.user_contributor, 2),
            (self.user_guest, 2),
            (self.user_viewer, 1),
            (self.user_no_roles, 1),
        ]
        url = (
            reverse('projectroles:search')
            + '?'
            + urlencode({'s': 'description'})
        )
        for user, count in expected:
            self.login_and_redirect(user, url, wait_elem=None, wait_loc='ID')
            elements = self.selenium.find_elements(
                By.CLASS_NAME, 'sodar-pr-project-list-item'
            )
            elem_set = set()
            for elem in elements:
                elem_set.add(elem.get_attribute('id'))
            self.assertEqual(len(elem_set), count)

    @override_settings(TIMELINE_SEARCH_LIMIT=2)
    def test_search_limit(self):
        """Test search limit"""
        url = reverse('projectroles:search') + '?' + urlencode({'s': 'event'})
        self.login_and_redirect(
            self.superuser, url, wait_elem=None, wait_loc='ID'
        )
        elements = self.selenium.find_elements(
            By.CLASS_NAME, 'sodar-pr-project-list-item'
        )
        self.assertEqual(len(elements), 2)



"""Utils tests for the projectroles app"""

import re

from django.conf import settings
from django.test import override_settings

from test_plus import TestCase

from projectroles.app_settings import AppSettingAPI
from projectroles.models import SODAR_CONSTANTS
from projectroles.utils import get_display_name, build_secret, get_app_names


app_settings = AppSettingAPI()


# SODAR constants
PROJECT_TYPE_CATEGORY = SODAR_CONSTANTS['PROJECT_TYPE_CATEGORY']
PROJECT_TYPE_PROJECT = SODAR_CONSTANTS['PROJECT_TYPE_PROJECT']

# Local constants
CONSTANTS_OVERRIDE = {
    'DISPLAY_NAMES': {
        'CATEGORY': {'default': 'bar', 'plural': 'bars'},
        'PROJECT': {'default': 'foo', 'plural': 'foos'},
    }
}


class TestUtils(TestCase):
    """Tests for general utilities"""

    def test_get_display_name(self):
        """Test get_display_name()"""
        self.assertEqual(get_display_name(PROJECT_TYPE_PROJECT), 'project')
        self.assertEqual(
            get_display_name(PROJECT_TYPE_PROJECT, title=True), 'Project'
        )
        self.assertEqual(
            get_display_name(PROJECT_TYPE_PROJECT, count=3), 'projects'
        )
        self.assertEqual(
            get_display_name(PROJECT_TYPE_PROJECT, title=True, count=3),
            'Projects',
        )
        self.assertEqual(get_display_name(PROJECT_TYPE_CATEGORY), 'category')
        self.assertEqual(
            get_display_name(PROJECT_TYPE_CATEGORY, title=True), 'Category'
        )
        self.assertEqual(
            get_display_name(PROJECT_TYPE_CATEGORY, count=3), 'categories'
        )
        self.assertEqual(
            get_display_name(PROJECT_TYPE_CATEGORY, title=True, count=3),
            'Categories',
        )

    # TODO: Test with override

    def test_build_secret(self):
        """Test build_secret()"""
        secret = build_secret()
        self.assertEqual(re.match(r'[a-z\d]{32}', secret).string, secret)
        self.assertEqual(len(build_secret(16)), 16)

    @override_settings(PROJECTROLES_SECRET_LENGTH=16)
    def test_build_secret_override(self):
        """Test build_secret() with default length setting override"""
        self.assertEqual(len(build_secret()), 16)

    def test_get_app_names(self):
        """Test get_app_names()"""
        app_names = get_app_names()
        self.assertNotEqual(len(app_names), 0)
        self.assertFalse(any([a.startswith('django.') for a in app_names]))
        self.assertFalse(any(['.apps.' in a for a in app_names]))
        self.assertNotIn(settings.SITE_PACKAGE, app_names)



"""
Removeroles management command for removing all roles from a user.
"""

import sys

from typing import Optional

from django.contrib.auth import get_user_model
from django.core.management.base import BaseCommand
from django.db import transaction

from projectroles.management.logging import ManagementCommandLogger
from projectroles.models import (
    Project,
    RoleAssignment,
    SODARUser,
    SODAR_CONSTANTS,
)
from projectroles.views import (
    RoleAssignmentOwnerTransferMixin,
    RoleAssignmentDeleteMixin,
)

logger = ManagementCommandLogger(__name__)
User = get_user_model()


# SODAR constants
PROJECT_ROLE_OWNER = SODAR_CONSTANTS['PROJECT_ROLE_OWNER']

# Local constants
CHECK_MODE_MSG = 'Check mode enabled, database will not be altered'
USER_NOT_FOUND_MSG = 'User not found with username: {}'


class Command(
    RoleAssignmentOwnerTransferMixin, RoleAssignmentDeleteMixin, BaseCommand
):
    help = (
        'Remove all roles from a user. Replace owner roles with given user or '
        'parent owner.'
    )

    @classmethod
    def _get_parent_owner(
        cls, project: Project, prev_owner: SODARUser
    ) -> Optional[SODARUser]:
        """Return assignment for first parent owner who is not previous owner"""
        if not project.parent:
            return None
        parent_owner_as = project.parent.get_owner()
        if parent_owner_as.user != prev_owner:
            return parent_owner_as.user
        return cls._get_parent_owner(project.parent, prev_owner)

    def _reassign_owner(
        self,
        project: Project,
        user: SODARUser,
        owner: Optional[SODARUser],
        role_as: RoleAssignment,
        p_title: str,
        check: bool,
    ) -> bool:
        """Reassign owner role"""
        # Fail top category if no new owner is specified
        if not owner and not project.parent:
            logger.warning(
                f'Unable to transfer ownership for top level '
                f'{project.type.lower()} {p_title}: no new owner provided'
            )
            return False
        # Get parent owner if not set
        if not owner:
            owner = self._get_parent_owner(project, user)
        # Fail if alternate parent owner is not found
        if not owner:
            logger.warning(
                f'Unable to transfer ownership in {p_title}: no parent owner '
                f'found'
            )
            return False
        if check:
            logger.info(
                f'Found role "{PROJECT_ROLE_OWNER}" in {p_title}: transfer to '
                f'user {owner.username}'
            )
            return True
        try:
            with transaction.atomic():
                self.transfer_owner(
                    project=project,
                    new_owner=owner,
                    old_owner_as=role_as,
                    old_owner_role=None,
                    notify_old=False,
                )
                logger.info(
                    f'Transferred ownership in {p_title} to {owner.username}'
                )
                return True
        except Exception as ex:
            logger.error(
                f'Failed to transfer ownership in {p_title} to '
                f'{owner.username}: {ex}'
            )
            return False

    def _remove_role(
        self, role_as: RoleAssignment, p_title: str, check: bool
    ) -> bool:
        """Remove non-owner role"""
        r_name = role_as.role.name
        if check:
            logger.info(f'Found role "{r_name}" in {p_title}: to be removed')
            return True
        try:
            with transaction.atomic():
                self.delete_assignment(role_as, None, False)
            logger.info(f'Deleted role "{r_name}" from {p_title}')
            return True
        except Exception as ex:
            logger.error(
                f'Failed to delete assignment "{r_name}" from {p_title}: {ex}'
            )
            return False

    def add_arguments(self, parser):
        parser.add_argument(
            '-u',
            '--user',
            dest='user',
            required=True,
            help='User name of user whose roles will be removed',
        )
        parser.add_argument(
            '-o',
            '--owner',
            dest='owner',
            required=False,
            help='Set owner role for user by given user name if set, otherwise '
            'set to parent owner',
        )
        parser.add_argument(
            '-d',
            '--deactivate',
            dest='deactivate',
            required=False,
            default=False,
            action='store_true',
            help='Deactivate user after removing their roles',
        )
        parser.add_argument(
            '-c',
            '--check',
            dest='check',
            required=False,
            default=False,
            action='store_true',
            help='Log roles to be removed or transferred without altering the '
            'database',
        )

    def handle(self, *args, **options):
        check = options.get('check', False)
        if check:
            logger.info(CHECK_MODE_MSG)
        if options['user'] == options.get('owner'):
            logger.error(
                'Same username given for both user and new owner: {}'.format(
                    options['user']
                )
            )
            sys.exit(1)
        try:
            user = User.objects.get(username=options['user'])
        except User.DoesNotExist:
            logger.error(USER_NOT_FOUND_MSG.format(options['user']))
            sys.exit(1)
        owner_name = options.get('owner')
        owner = None
        if owner_name:
            try:
                owner = User.objects.get(username=owner_name)
            except User.DoesNotExist:
                logger.error(USER_NOT_FOUND_MSG.format(owner_name))
                sys.exit(1)

        logger.info(
            '{} roles {} user "{}"..'.format(
                'Checking' if check else 'Removing',
                'for' if check else 'from',
                user.username,
            )
        )
        if owner:
            logger.info(
                f'New owner for replacing owner roles: {owner.username}'
            )
        role_count = 0
        fail_count = 0
        roles = RoleAssignment.objects.filter(user=user).order_by(
            'project__full_title'
        )
        if roles.count() == 0:
            logger.info('No roles found')
            return

        for role_as in roles:
            project = role_as.project
            p_title = project.get_log_title()
            if project.is_remote():  # Skip remote projects
                logger.debug(f'Skipping remote project: {p_title}')
                continue
            # Owner role reassignment
            if role_as.role.name == PROJECT_ROLE_OWNER:
                update_ok = self._reassign_owner(
                    project, user, owner, role_as, p_title, check
                )
            else:  # Non-owner role removal
                update_ok = self._remove_role(role_as, p_title, check)
            if not check and update_ok is True:
                role_count += 1
            elif not check:
                fail_count += 1
        if options['deactivate']:
            if check:
                logger.info(f'User to be deactivated: {user.username}')
            else:
                user.is_active = False
                user.save()
                logger.info(f'User "{user.username}" deactivated')
        if check:
            logger.info('Check done')
        else:
            logger.info(
                f'Removed roles from user "{user.username}" ({role_count} OK, '
                f'{fail_count} failed)'
            )



"""UI views for the backgroundjobs app"""

from django.conf import settings
from django.contrib import messages
from django.shortcuts import redirect
from django.urls import reverse
from django.views.generic import TemplateView, ListView

from bgjobs.models import BackgroundJob

# Projectroles dependency
from projectroles.views import (
    LoginRequiredMixin,
    LoggedInPermissionMixin,
    ProjectContextMixin,
    ProjectPermissionMixin,
)
from projectroles.plugins import PluginAPI


plugin_api = PluginAPI()


# Local variables
DEFAULT_PAGINATION = 15


class SiteBackgroundJobView(
    LoggedInPermissionMixin,
    ListView,
):
    """Site-wide background jobs view"""

    permission_required = 'bgjobs.view_site_bgjobs'
    template_name = 'bgjobs/site_backgroundjobs.html'
    model = BackgroundJob
    paginate_by = getattr(settings, 'BGJOBS_PAGINATION', DEFAULT_PAGINATION)

    def get_queryset(self):
        return super().get_queryset().filter(project__isnull=True)


class ProjectBackgroundJobView(
    LoginRequiredMixin,
    LoggedInPermissionMixin,
    ProjectPermissionMixin,
    ProjectContextMixin,
    ListView,
):
    """Project background jobs view"""

    template_name = 'bgjobs/project_backgroundjobs.html'
    permission_required = 'bgjobs.view_jobs_own'
    model = BackgroundJob
    paginate_by = getattr(settings, 'BGJOBS_PAGINATION', DEFAULT_PAGINATION)

    def get_queryset(self):
        # TODO: filter to user's job if can only see their own
        return (
            super()
            .get_queryset()
            .filter(project__sodar_uuid=self.kwargs['project'])
        )


class BackgroundJobClearViewBase(
    LoginRequiredMixin,
    LoggedInPermissionMixin,
    ProjectPermissionMixin,
    ProjectContextMixin,
    TemplateView,
):
    """Base class for view clearing jobs"""

    #: The template is the same for both sub classes
    template_name = 'bgjobs/backgroundjob_confirm_clear.html'
    #: Set in sub class
    which_jobs = None
    #: Set in sub class
    permission_required = None

    def get_context_data(self, *args, **kwargs):
        result = super().get_context_data(*args, **kwargs)
        result['which_jobs'] = self.which_jobs
        return result

    def post(self, _request, **_kwargs):
        """Handle form POST."""
        context = self.get_context_data()
        project = context['project']
        filter_kwargs = {'project': project}
        if self.which_jobs != 'all':
            filter_kwargs['user'] = self.request.user

        try:
            bg_jobs = BackgroundJob.objects.filter(**filter_kwargs)
            bg_job_count = bg_jobs.count()
            bg_jobs.delete()

            timeline = plugin_api.get_backend_api('timeline_backend')
            if timeline:
                timeline.add_event(
                    project=self.get_project(self.request, self.kwargs),
                    app_name='bgjobs',
                    user=self.request.user,
                    event_name='clear_bg_jobs',
                    description='Clearing {} background jobs'.format(
                        'user-owned' if self.which_jobs != 'all' else 'all'
                    ),
                    status_type=timeline.TL_STATUS_OK,
                )
            messages.success(
                self.request, f'Removed {bg_job_count} background jobs.'
            )
        except Exception as ex:
            messages.error(
                self.request, f'Unable to remove background jobs: {ex}'
            )
        return redirect(
            reverse('bgjobs:list', kwargs={'project': project.sodar_uuid})
        )


class BackgroundJobClearOwnView(BackgroundJobClearViewBase):
    """View for clearing a user's own background job"""

    which_jobs = 'own'
    permission_required = 'bgjobs.update_bgjob_own'


class BackgroundJobClearAllView(BackgroundJobClearViewBase):
    """View for clearing a background jobs in a project"""

    which_jobs = 'all'
    permission_required = 'bgjobs.update_bgjob_all'
